import torch
import torch.nn as nn


class PearsonLoss(nn.Module):
    def __init__(self, **kwargs):
        super(PearsonLoss, self).__init__()

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        mean_x = torch.mean(x)
        mean_y = torch.mean(y)
        xm = x.sub(mean_x)
        ym = y.sub(mean_y)
        r_num = xm.dot(ym)
        r_den = torch.norm(xm, 2) * torch.norm(ym, 2)
        r_den = torch.clamp(r_den, min=1e-6)
        r_val = r_num / r_den
        r_val = torch.clamp(r_val, min=-1.0, max=1.0)
        return 1 - r_val



"""Refiner to summarize the query/community"""

import logging
from abc import ABC, abstractmethod
from typing import Optional, Union

from .generator import BaseLLM
from .prompts import SUMMARIZE_PROMPT

logger = logging.getLogger(__name__)

DEFAULT_MAX_INPUT_TOKENS = 4000
DEFAULT_MAX_SUMMARY_LENGTH = 500


class BaseRefiner(ABC):
    @abstractmethod
    def refine(self, context):
        """Refine the context"""


class LLMRefiner(BaseRefiner):
    def __init__(
        self,
        llm: BaseLLM,
        prompt: Optional[str] = None,
        max_input_tokens: Optional[int] = None,
        max_summary_length: Optional[int] = None,
    ):
        self.llm = llm
        self.prompt = prompt or SUMMARIZE_PROMPT
        self.max_input_tokens = max_input_tokens or DEFAULT_MAX_INPUT_TOKENS
        self.max_summary_length = max_summary_length or DEFAULT_MAX_SUMMARY_LENGTH

    def refine(self, context):
        return self.llm.generate(self.prompt)



"""Retrieve pipeline
retrieve: one query match with top_k document: [{'id': 'doc_id', 'similarity': 0.8}, {'id': 'doc_id', "similarity': 0.9]
"""

import glob
import logging
import os
import pickle
from argparse import ArgumentParser

import numpy as np

from ..models.retrieval_auto import AutoModelForRetrieval, FaissRetrieval

logger = logging.getLogger(__name__)
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
    level=logging.INFO,
)


def load_pickle(path):
    with open(path, 'rb') as f:
        reps, lookup = pickle.load(f)
    return np.array(reps), lookup


def save_ranking(dists, indices, ranking_file, query_ids):
    """
    save format: query_id, doc_id, score, rank
    """
    os.makedirs(os.path.dirname(ranking_file), exist_ok=True)

    with open(ranking_file, 'w') as f:
        for qid, score, index in zip(query_ids, dists, indices):
            score_list = [(s, idx) for s, idx in zip(score, index)]
            score_list = sorted(score_list, key=lambda x: x[0], reverse=True)
            for rank, (s, idx) in enumerate(score_list, start=1):
                f.write(f'{qid}\t{idx}\t{s}\t{rank}\n')


def retrieve():
    parser = ArgumentParser()
    parser.add_argument("--query_reps", required=True)
    parser.add_argument("--passage_reps", required=True)
    parser.add_argument("--top_k", type=int, default=100)
    parser.add_argument("--batch_size", type=int, default=128)
    parser.add_argument("--save_ranking_file", required=True)
    parser.add_argument("--save_text", action="store_true")

    args = parser.parse_args()

    index_files = glob.glob(args.passage_reps)
    logger.info(f"Pattern match found {len(index_files)} files for corpus; loading them into index.")

    passage_reps0, passage_lookup0 = load_pickle(index_files[0])
    logger.info(f"Loaded info vector embedding of shape {passage_reps0.shape[1]}")
    retriever = FaissRetrieval(passage_reps0)

    shards = [(passage_reps0, passage_lookup0)]
    for i in range(1, len(index_files)):
        passage_reps, passage_lookup = load_pickle(index_files[i])
        shards.append((passage_reps, passage_lookup))

    lookups = []
    for passage_reps, passage_lookup in shards:
        retriever.add(passage_reps)
        lookups += passage_lookup

    query_reps, query_lookup = load_pickle(args.query_reps)

    dists, indices = retriever.search(query_embeddings=query_reps, top_k=args.top_k)
    doc_ids = np.array([[int(lookups[idx]) for idx in indices] for indices in indices])

    save_ranking(dists, doc_ids, args.save_ranking_file, query_lookup)


if __name__ == "__main__":
    retrieve()



"""
trainable temperature parameter: Text and Code Embeddings by Contrastive Pre-Training
"""

import logging

import torch
from torch import nn

logger = logging.getLogger(__name__)


class CosineSimilarity(nn.Module):
    def __init__(self, temperature: float = 0.05, dynamic_temperature=False, **kwargs):
        """
        Temperature should be smaller than 1 when use cosine similarity (normalized=True). Recommend to set it 0.01-0.1
        """
        super().__init__()
        self.temperature = temperature
        self.dynamic_temperature = dynamic_temperature

    def forward(self, query_embeddings: torch.Tensor, document_embeddings: torch.Tensor):
        sim_pos_vector = torch.cosine_similarity(query_embeddings, document_embeddings, dim=-1)
        sim_pos_vector = sim_pos_vector / self.temperature
        sim_neg_matrix = torch.cosine_similarity(
            query_embeddings.unsqueeze(1),
            document_embeddings.unsqueeze(0),
            dim=-1,
        )
        sim_neg_matrix = sim_neg_matrix / self.temperature
        sim_diff_matrix = sim_pos_vector.unsqueeze(1) - sim_neg_matrix
        loss = -torch.log(torch.sigmoid(sim_diff_matrix)).mean()
        return loss

    def get_temperature(self):
        if not self.dynamic_temperature:
            return self.temperature
        return torch.clamp(self.temperature, min=1e-3)


class Similarity(nn.Module):
    """
    Dot product or cosine similarity
    """

    def __init__(self, temp=0.05):
        super().__init__()
        self.temp = temp
        self.cos = nn.CosineSimilarity(dim=-1)

    def forward(self, x, y):
        return self.cos(x, y) / self.temp


class ContrastiveLoss(nn.Module):
    def __init__(self, margin=0.5):
        super(ContrastiveLoss, self).__init__()
        self.margin = margin

    def forward(self, pred: torch.Tensor, label: torch.Tensor):
        distance = 1.0 - pred
        loss = label.float() * distance.pow(2) + (1 - label).float() * torch.clamp(self.margin - distance, min=0).pow(2)

        return loss.sum()



"""Integration with Langchain"""

import logging
from typing import Any, Dict, List, Optional, Sequence

from langchain.llms.base import LLM
from langchain.retrievers.document_compressors.base import BaseDocumentCompressor
from langchain_core.callbacks.manager import (
    CallbackManagerForLLMRun,
    CallbackManagerForRetrieverRun,
    Callbacks,
)
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.retrievers import BaseRetriever
from transformers import AutoModelForCausalLM, AutoTokenizer

from ..models.embedding_auto import AutoModelForEmbedding
from ..models.rerank import AutoModelForRanking
from .generator import BaseLLM

logger = logging.getLogger(__name__)


class LangchainEmbedding(AutoModelForEmbedding, Embeddings):
    """Embedding integrated with Langchain
    Example:
        .. code-block:: python

            from retrievals import LangchainEmbedding

            model_name_or_path = "BAAI/bge-large-en"
            model_kwargs = {'device': 'cpu'}
            encode_kwargs = {'normalize_embeddings': True}
            hf = LangchainEmbedding(
                model_name_or_path=model_name_or_path,
                model_kwargs=model_kwargs,
                encode_kwargs=encode_kwargs
            )
    """

    client: Any
    model_name: Optional[str] = None
    cache_folder: Optional[str] = None
    model_kwargs: Dict[str, Any] = dict()
    encode_kwargs: Dict[str, Any] = dict()

    def __init__(self, model_name=None, **model_kwargs):
        Embeddings.__init__(self)
        self.model_kwargs = model_kwargs

        model = AutoModelForEmbedding.from_pretrained(model_name_or_path=model_name, **model_kwargs)
        for key, value in model.__dict__.items():
            self.__dict__[key] = value

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a HuggingFace transformer model."""
        embeddings = self.encode(texts, **self.encode_kwargs)
        return embeddings.tolist()

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a HuggingFace transformer model."""
        return self.embed_documents([text])[0]


class LangchainRetrieval(BaseRetriever):
    model: Any
    kwargs: dict = dict()

    def _get_relevant_documents(
        self,
        query: str,
        *,
        run_manager: CallbackManagerForRetrieverRun,  # noqa
    ) -> List[Document]:
        """Get documents relevant to a query."""
        docs = self.model.search(query, **self.kwargs)
        return [Document(page_content=doc["content"], metadata=doc.get("document_metadata", {})) for doc in docs]


class LangchainReranker(BaseDocumentCompressor):
    model: Any
    kwargs: dict = dict()
    k: int = 5

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def compress_documents(
        self, documents: Sequence[Document], query: str, callbacks: Optional[Callbacks] = None, **kwargs
    ) -> Sequence[Document]:
        """
        Compress documents using `BCEmbedding RerankerModel API`.

        Args:
            documents: A sequence of documents to compress.
            query: The query to use for compressing the documents.
            callbacks: Callbacks to run during the compression process.

        Returns:
            A sequence of compressed documents.
        """
        if len(documents) == 0:
            return []
        doc_list = list(documents)

        passages = []
        valid_doc_list = []
        invalid_doc_list = []
        for d in doc_list:
            passage = d.page_content
            if isinstance(passage, str) and len(passage) > 0:
                passages.append(passage.replace('\n', ' '))
                valid_doc_list.append(d)
            else:
                invalid_doc_list.append(d)

        rerank_result = self.model.rerank(query, passages)
        final_results = []
        for score, doc_id in zip(rerank_result['rerank_scores'], rerank_result['rerank_ids']):
            doc = valid_doc_list[doc_id]
            doc.metadata["relevance_score"] = score
            final_results.append(doc)
        for doc in invalid_doc_list:
            doc.metadata["relevance_score"] = 0
            final_results.append(doc)

        final_results = final_results[: self.top_n]
        return final_results


class LangchainLLM(LLM):
    tokenizer: AutoTokenizer = None
    model: AutoModelForCausalLM = None
    max_tokens: int = 2048
    temperature: float = 0.1
    top_p: float = 0.9
    history: List[str] = []

    def __init__(
        self,
        model_name_or_path: str,
        trust_remote_code: bool = True,
        temperature: float = 0.1,
        max_tokens: int = 2048,
        top_p: float = 0.9,
        **kwargs: Any,
    ):
        super(LangchainLLM, self).__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code)
        if self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({'pad_token': '[PAD]'})
        self.model = (
            AutoModelForCausalLM.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code).half().cuda()
        )
        self.model = self.model.eval()
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.top_p = top_p

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ):
        if hasattr(self.model, 'chat') and callable(self.model.chat):
            response, _ = self.model.chat(self.tokenizer, prompt, history=self.history)
            self.history = self.history + [[None, response]]
            return response
        else:
            batch_inputs = self.tokenizer.batch_encode_plus([prompt], padding='longest', return_tensors='pt')
            batch_inputs["input_ids"] = batch_inputs["input_ids"].cuda()
            batch_inputs["attention_mask"] = batch_inputs["attention_mask"].cuda()

            output = self.model.generate(
                **batch_inputs,
                max_new_tokens=self.max_tokens,
                do_sample=True,
                temperature=self.temperature,
                top_p=self.top_p,
            )
            response = self.tokenizer.batch_decode(
                output.cpu()[:, batch_inputs["input_ids"].shape[-1] :], skip_special_tokens=True
            )[0]

            self.history = self.history + [[None, response]]
            return response

    @property
    def _llm_type(self):
        return "Open-retrievals-llm"



import torch
import torch.nn.functional as F
from torch import nn


class BCELoss(nn.Module):
    def __init__(self, temperature: float = 1.0):
        super().__init__()
        self.temperature = temperature

    def forward(self, inputs, labels, mask=None, sample_weight=None, class_weight=None):
        bce1 = F.binary_cross_entropy(inputs, torch.ones_like(inputs), reduction="none")
        bce2 = F.binary_cross_entropy(inputs, torch.zeros_like(inputs), reduction="none")
        bce = 1 * bce1 * labels + bce2 * (1 - labels)
        if mask is not None:
            # mask = torch.where(targets >= 0, torch.ones_like(bce), torch.zeros_like(bce))
            bce = bce * mask

        if sample_weight is not None:
            bce = bce * sample_weight.unsqueeze(1)
        loss = torch.sum(bce, dim=1)  # / torch.sum(mask, dim=1)
        loss = loss.mean()
        return loss



import torch


def get_cos_sim(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """
    Computes the cosine similarity cos_sim(a[i], b[j]) for all i and j.

    :return: Matrix with res[i][j]  = cos_sim(a[i], b[j])
    """
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)

    if not isinstance(b, torch.Tensor):
        b = torch.tensor(b)

    if len(a.shape) == 1:
        a = a.unsqueeze(0)

    if len(b.shape) == 1:
        b = b.unsqueeze(0)

    a_norm = torch.nn.functional.normalize(a, p=2, dim=1)
    b_norm = torch.nn.functional.normalize(b, p=2, dim=1)
    return torch.mm(a_norm, b_norm.transpose(0, 1))



import logging
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

import torch
import torch.distributed as dist
import torch.nn as nn
from transformers import PreTrainedTokenizer

logger = logging.getLogger(__name__)


class Base(ABC, torch.nn.Module):
    """Base class for embedding and reranking model"""

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        **kwargs,
    ):
        super().__init__()
        if isinstance(model, str):
            assert ValueError("Please use AutoModelForEmbedding.from_pretrained(model_name_or_path)")
        self.model = model
        self.tokenizer = tokenizer

    @abstractmethod
    def forward(self, *args, **kwargs):
        """Pytorch forward method."""

    def save_pretrained(self, path: str, safe_serialization: bool = True):
        """Saves all model and tokenizer to path"""
        logger.info("Save model to {}".format(path))
        state_dict = self.model.state_dict()
        state_dict = type(state_dict)({k: v.clone().cpu() for k, v in state_dict.items()})
        self.model.save_pretrained(path, state_dict=state_dict, safe_serialization=safe_serialization)
        self.tokenizer.save_pretrained(path)

    def gradient_checkpointing_enable(self, gradient_checkpointing_kwargs=None):
        self.model.gradient_checkpointing_enable(gradient_checkpointing_kwargs=gradient_checkpointing_kwargs)

    def resize_token_embeddings(self, new_num_tokens: Optional = None, pad_to_multiple_of: Optional = None):
        # add new, random embeddings for the new tokens
        self.model.resize_token_embeddings(new_num_tokens, pad_to_multiple_of)

    def push_to_hub(self, hub_model_id: str, private: bool = True, **kwargs):
        """push model to hub

        :param hub_model_id: str, hub model id.
        :param private: bool, whether push to private repo. Default True.
        :param kwargs: other kwargs for `push_to_hub` method.
        """
        self.tokenizer.push_to_hub(hub_model_id, private=private, **kwargs)
        self.backbone.push_to_hub(hub_model_id, private=private, **kwargs)

    def _dist_gather_tensor(self, tensor: Optional[torch.Tensor]):
        if tensor is None:
            return None
        tensor = tensor.contiguous()

        all_tensors = [torch.empty_like(tensor) for _ in range(self.world_size)]
        dist.all_gather(all_tensors, tensor)

        all_tensors[self.process_rank] = tensor
        all_tensors = torch.cat(all_tensors, dim=0)
        return all_tensors

    def _text_length(self, text: Union[List[int], List[List[int]]]):
        """
        Help function to get the length for the input text. Text can be either
        a list of ints (which means a single text as input), or a tuple of list of ints
        (representing several text inputs to the model).
        """

        if isinstance(text, dict):  # {key: value} case
            return len(next(iter(text.values())))
        elif not hasattr(text, "__len__"):  # Object has no len() method
            return 1
        elif len(text) == 0 or isinstance(text[0], int):  # Empty string or list of ints
            return len(text)
        else:
            return sum([len(t) for t in text])  # Sum of length of individual strings



"""Rewriter of query"""

import logging
from abc import ABC, abstractmethod

from .generator import BaseLLM

logger = logging.getLogger(__name__)


class BaseRewriter(ABC):
    @abstractmethod
    def rewrite(self, query: str) -> str:
        """Rewrite the query"""


class HyDE(BaseRewriter):
    """
    https://github.com/texttron/hyde/blob/main/src/hyde/hyde.py
    """

    def __init__(self, llm: BaseLLM, prompt: str):
        self.llm = llm
        self.prompt = prompt

    def rewrite(self, query: str):
        return self.llm.generate(self.prompt)



import random
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from transformers import (
    BatchEncoding,
    DataCollatorForSeq2Seq,
    DataCollatorWithPadding,
    PreTrainedTokenizer,
)


class AutoCollator(DataCollatorWithPadding):
    """Choose the collator based on data/task"""

    def __init__(self):
        pass


class PairCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        query_max_length: int = 32,
        document_max_length: int = 128,
        append_eos_token: bool = False,
        query_key: str = 'query',
        document_key: str = 'positive',
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_max_length = query_max_length
        self.document_max_length = document_max_length
        self.query_key = query_key
        self.document_key = document_key

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        # TODO: tokenizer.apply_chat_template(chat, tokenize=False)
        assert len(features) > 0
        if isinstance(features[0], dict):
            assert (
                self.query_key in features[0] and self.document_key in features[0]
            ), f"PairCollator should have {self.query_key} and {self.document_key} in features, while get {features[0]}"

            query_texts = [feature[self.query_key] for feature in features]
            document_texts = [feature[self.document_key] for feature in features]
        elif isinstance(features[0], (list, tuple)):
            query_texts = [f[0] for f in features]
            document_texts = [f[1] for f in features]
        else:
            raise ValueError

        if isinstance(query_texts[0], list):
            query_texts = sum(query_texts, [])
        if isinstance(document_texts[0], list):
            document_texts = sum(document_texts, [])  # flatten nested list

        if isinstance(query_texts[0], str):
            tokenize_fn = self.tokenizer
            tokenize_args = {
                "truncation": True,
            }
        else:
            tokenize_fn = self.tokenizer.pad
            tokenize_args = {
                "pad_to_multiple_of": None,
            }

        query_inputs = tokenize_fn(
            query_texts, padding="max_length", max_length=self.query_max_length, return_tensors="pt", **tokenize_args
        )
        document_inputs = tokenize_fn(
            document_texts,
            padding="max_length",
            max_length=self.document_max_length,
            return_tensors="pt",
            **tokenize_args,
        )

        return {self.query_key: query_inputs, self.document_key: document_inputs}


class TripletCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        query_max_length: int = 32,
        document_max_length: int = 128,
        append_eos_token: bool = False,
        query_key: str = 'query',
        positive_key: str = 'positive',
        negative_key: Optional[str] = 'negative',
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_max_length = query_max_length
        self.document_max_length = document_max_length

        self.query_key = query_key
        self.positive_key = positive_key
        self.negative_key = negative_key

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        assert len(features) > 0
        if isinstance(features[0], dict):
            assert (
                self.positive_key in features[0]
                and self.positive_key in features[0]
                and self.negative_key in features[0]
            ), f"TripletCollator should have {self.query_key}, {self.positive_key} and {self.negative_key} in dict key"

            query_texts = [feature[self.query_key] for feature in features]
            pos_texts = [feature[self.positive_key] for feature in features]
            neg_texts = [feature[self.negative_key] for feature in features]
        elif isinstance(features[0], (list, tuple)):
            query_texts = [feature[0] for feature in features]
            pos_texts = [feature[1] for feature in features]
            neg_texts = [feature[2] for feature in features]
        else:
            raise ValueError

        if isinstance(query_texts[0], list):
            query_texts = sum(query_texts, [])
        if isinstance(pos_texts[0], list):
            pos_texts = sum(pos_texts, [])
        if isinstance(neg_texts[0], list):
            neg_texts = sum(neg_texts, [])

        if isinstance(query_texts[0], str):
            tokenize_fn = self.tokenizer
            tokenize_args = {
                "truncation": True,
                "return_token_type_ids": False,
                "add_special_tokens": True,
            }
        else:
            tokenize_fn = self.tokenizer.pad
            tokenize_args = {
                "pad_to_multiple_of": None,
            }

        query_inputs = tokenize_fn(
            query_texts, padding="max_length", max_length=self.query_max_length, return_tensors="pt", **tokenize_args
        )
        pos_inputs = tokenize_fn(
            pos_texts, padding="max_length", max_length=self.document_max_length, return_tensors="pt", **tokenize_args
        )
        neg_inputs = tokenize_fn(
            neg_texts, padding="max_length", max_length=self.document_max_length, return_tensors="pt", **tokenize_args
        )

        return {
            self.query_key: query_inputs,
            self.positive_key: pos_inputs,
            self.negative_key: neg_inputs,
        }


class RerankCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        max_length: int = 128,
        append_eos_token: bool = False,
        query_key: str = 'query',
        document_key: str = 'document',
    ):
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.max_length = max_length
        self.query_key = query_key
        self.document_key = document_key

    def __call__(self, features: Union[List[Dict[str, Any]], List]) -> BatchEncoding:
        assert len(features) > 0
        if isinstance(features[0], dict):
            assert (
                self.query_key in features[0] and self.document_key in features[0]
            ), f"RerankCollator should have '{self.query_key}' and '{self.document_key}' keys in features, maybe labels"
            query_texts = [feature[self.query_key] for feature in features]
            document_texts = [feature[self.document_key] for feature in features]
        else:
            query_texts = [feature[0] for feature in features]
            document_texts = [feature[1] for feature in features]

        if isinstance(query_texts[0], str):
            tokenize_fn = self.tokenizer
            tokenize_args = {
                "truncation": True,
            }
        else:
            tokenize_fn = self.tokenizer.pad
            tokenize_args = {
                "pad_to_multiple_of": None,
            }

        batch = tokenize_fn(
            text=query_texts,
            text_pair=document_texts,
            padding="max_length",
            max_length=self.max_length,
            return_tensors="pt",
            **tokenize_args,
        )

        if 'labels' in features[0].keys():
            labels = [feature['labels'] for feature in features]
            batch['labels'] = torch.tensor(labels, dtype=torch.float32)
        return batch


class ColBertCollator(DataCollatorWithPadding):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        query_max_length: int = 32,
        document_max_length: int = 128,
        query_key: str = 'query',
        positive_key: str = 'positive',
        negative_key: str = 'negative',
    ) -> None:
        self.tokenizer = tokenizer
        if not hasattr(self.tokenizer, "pad_token_id") or self.tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": "[PAD]"})

        self.query_max_length = query_max_length
        self.document_max_length = document_max_length
        self.query_key = query_key
        self.positive_key = positive_key
        self.negative_key = negative_key

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        assert len(features) > 0
        assert (
            self.query_key in features[0] and self.positive_key in features[0]
        ), f"PairCollator should have '{self.query_key}' and '{self.positive_key}' in features, while get {features[0]}"

        query_texts = [feature[self.query_key] for feature in features]
        pos_texts = [feature[self.positive_key] for feature in features]

        if isinstance(query_texts[0], list):
            query_texts = sum(query_texts, [])
        if isinstance(pos_texts[0], list):
            pos_texts = sum(pos_texts, [])

        if isinstance(query_texts[0], str):
            tokenize_fn = self.tokenizer
            tokenize_args = {
                "truncation": True,
            }
        else:
            tokenize_fn = self.tokenizer.pad
            tokenize_args = {
                "pad_to_multiple_of": None,
            }

        query_inputs = tokenize_fn(
            query_texts, padding="max_length", max_length=self.query_max_length, return_tensors="pt", **tokenize_args
        )
        pos_inputs = tokenize_fn(
            pos_texts, padding="max_length", max_length=self.document_max_length, return_tensors="pt", **tokenize_args
        )

        batch = {
            'query_input_ids': query_inputs['input_ids'],
            'query_attention_mask': query_inputs['attention_mask'],
            'pos_input_ids': pos_inputs['input_ids'],
            'pos_attention_mask': pos_inputs['attention_mask'],
        }

        if self.negative_key in features[0]:
            neg_texts = [feature[self.negative_key] for feature in features]

            if isinstance(neg_texts[0], list):
                neg_texts = sum(neg_texts, [])  # flatten nested list

            neg_inputs = tokenize_fn(
                neg_texts,
                padding='max_length',
                max_length=self.document_max_length,
                return_tensors='pt',
                **tokenize_args,
            )
            batch.update({'neg_input_ids': neg_inputs['input_ids'], 'neg_attention_mask': neg_inputs['attention_mask']})

        return batch


class LLMRerankCollator(DataCollatorForSeq2Seq):
    """Rerank collator for causal llm, with examples query, positive and negative"""

    tokenizers: PreTrainedTokenizer
    query_key: str = 'query'
    positive_key: str = 'positive'
    negative_key: str = 'negative'
    max_length: int = 128
    query_instruction: Optional[str] = None
    document_instruction: Optional[str] = None

    def __init__(self, tokenizer: PreTrainedTokenizer, prompt: str, max_length: int = 128):
        self.tokenizer = tokenizer
        self.prompt = prompt
        self.max_length = max_length

    def __call__(self, features: List[Dict[str, Any]], return_tensors='pt'):
        examples = []

        if isinstance(features[0], dict):
            for i in range(len(features)):
                examples.append((features[i][self.query_key], features[i][self.positive_key]))
                for neg in features[i][self.negative_key]:
                    examples.append((features[i][self.query_key], neg))
        else:
            examples = features

        batch = self.tokenizer(
            [i[0] for i in examples],
            ["\n" + i[1] + '\n' + self.prompt + 'Yes' for i in examples],
            return_tensors=None,
            max_length=self.max_length,
            truncation='only_second',
            add_special_tokens=False,
            padding=False,
            return_attention_mask=False,
            return_token_type_ids=False,
        )

        batch['attention_mask'] = [[1] * len(example) for example in batch['input_ids']]
        batch['labels'] = batch['input_ids'].copy()
        batch['labels'] = [[-100] * (len(example) - 1) + example[-1:] for example in batch['labels']]

        max_label_length = max(len(l) for l in batch['labels'])
        padding_side = self.tokenizer.padding_side
        for i in range(len(batch['labels'])):
            feature = batch['labels'][i]
            remainder = [self.label_pad_token_id] * (max_label_length - len(feature))
            if isinstance(feature, list):
                batch['labels'][i] = feature + remainder if padding_side == "right" else remainder + feature
            elif padding_side == "right":
                batch['labels'][i] = np.concatenate([feature, remainder]).astype(np.int64)
            else:
                batch['labels'][i] = np.concatenate([remainder, feature]).astype(np.int64)

        batch = self.tokenizer.pad(
            batch,
            padding='longest',
            max_length=self.max_length,
            return_tensors=return_tensors,
            pad_to_multiple_of=self.pad_to_multiple_of,
        )

        return batch


class EncodeCollator(DataCollatorWithPadding):
    def __init__(self, tokenizer: PreTrainedTokenizer, id_key: Optional[str] = None, **kwargs):
        self.tokenizer = tokenizer
        self.id_key = id_key

    def __call__(self, features):
        if self.id_key is not None:
            text_ids = [x[0] for x in features]
            text_features = [x[1] for x in features]
            collated_features = super().__call__(text_features)
            return text_ids, collated_features
        else:
            text_features = features
            collated_features = super().__call__(text_features)
            return collated_features


def mask_pad_token(q: Dict[str, torch.Tensor], prob=0.9):
    if random.random() > prob:
        tensor = q['input_ids'].float()
        mask = torch.rand(tensor.shape)
        mask = (mask > prob).float()
        tensor = tensor * (1 - mask) + 2 * mask
        tensor = tensor.long()
        q['input_ids'] = tensor
    return q



"""Tuning API for embedding, reranking, and LLM"""


class AutoTuner(object):
    """Given the data and model, assign the tuning type, then fine-tune for embedding, ranking and llm"""

    def __init__(self):
        pass

    def run(self):
        pass



import logging
import os
from typing import Any, Callable, Dict, List, Optional

import torch
import torch.distributed as dist
import torch.nn as nn
from transformers import Trainer

logger = logging.getLogger(__name__)


class RetrievalTrainer(Trainer):
    def __init__(
        self,
        loss_fn: Optional[Callable] = None,
        train_type: str = 'normal',
        negatives_x_device: bool = False,
        *args,
        **kwargs,
    ):
        super(RetrievalTrainer, self).__init__(*args, **kwargs)
        self.loss_fn = loss_fn
        self.train_type = train_type
        self._dist_loss_scale_factor = dist.get_world_size() if negatives_x_device else 1

    def training_step(self, *args):
        return super().training_step(*args) / self._dist_loss_scale_factor

    def compute_loss(self, model, inputs, return_outputs=False):
        if self.train_type == 'pairwise':
            return self.compute_pair_loss(model=model, inputs=inputs, return_outputs=return_outputs)

        outputs = model(inputs, return_dict=True)
        if isinstance(outputs, dict) and 'loss' in outputs:
            return outputs['loss']
        else:
            return self.loss_fn(*outputs)

    def compute_pair_loss(self, model: nn.Module, inputs: Any, return_outputs: bool = False):
        query = inputs["query"]
        if 'document' in inputs:
            pos = inputs["document"]
        else:
            pos = inputs['positive']
        query_embeddings = model(query)
        pos_embeddings = model(pos)
        if 'negative' in inputs:
            neg = inputs["negative"]
            neg_embeddings = model(neg)
        else:
            neg_embeddings = None

        loss = self.loss_fn(query_embeddings, pos_embeddings, neg_embeddings)
        if not return_outputs:
            return loss

        outputs = dict()
        outputs['query'] = query_embeddings
        outputs['positive'] = pos_embeddings
        if 'negative' in inputs:
            outputs['negative'] = neg_embeddings
        return loss, outputs

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info(f"Saving retrieval model checkpoint to {output_dir}")
        save_model = self.model.model if hasattr(self.model, 'model') else self.model
        save_model.save_pretrained(output_dir)
        self.model.tokenizer.save_pretrained(output_dir)


class RerankTrainer(Trainer):
    def __init__(self, loss_fn: Callable = None, **kwargs):
        super().__init__(**kwargs)
        if not loss_fn:
            loss_fn = nn.BCEWithLogitsLoss()
        self.loss_fn = loss_fn

    def compute_loss(self, model: nn.Module, inputs, return_outputs=False, **kwargs):
        outputs: dict = model(**inputs)
        loss = outputs['loss']
        return (loss, outputs) if return_outputs else loss

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info(f"Saving rerank model checkpoint to {output_dir}")
        self.model.save_pretrained(output_dir)
        self.model.tokenizer.save_pretrained(output_dir)


class DistilTrainer(Trainer):
    # https://github.com/texttron/tevatron/blob/tevatron-v1/src/tevatron/distillation/trainer.py
    def __init__(self, teacher_model):
        super(DistilTrainer, self).__init__()
        self.teacher_model = teacher_model
        self._dist_loss_scale_factor = 1

    def compute_loss(self, model, inputs, return_outputs=False):
        student_scores = model(inputs)
        with torch.no_grad():
            teacher_scores = self.teacher_model(inputs)

        teacher_mat = torch.zeros(student_scores.shape, dtype=student_scores.dtype, device=teacher_scores.device)
        index = torch.arange(teacher_scores.size(0), device=teacher_scores.device)
        teacher_scores = torch.softmax(
            teacher_scores.view(student_scores.size(0), -1), dim=1, dtype=student_scores.dtype
        )
        teacher_mat = torch.scatter(
            teacher_mat, dim=-1, index=index.view(student_scores.size(0), -1), src=teacher_scores
        )
        student_scores = nn.functional.log_softmax(student_scores, dim=1)
        loss = nn.functional.kl_div(student_scores, teacher_mat, reduction='batchmean') * self._dist_loss_scale_factor
        return loss



import gc
import logging
from typing import Any, Generic, List, Optional, TypeVar, cast

from .workflow import (
    PipelineWorkflowConfig,
    PipelineWorkflowStep,
    Workflow,
    WorkflowToRun,
)

logger = logging.getLogger(__name__)


def run_pipeline(workflows: List):
    loaded_workflows = load_workflows(workflows)
    workflows_to_run = loaded_workflows.workflows

    for workflow_to_run in workflows_to_run:
        gc.collect()

        workflow = workflow_to_run.workflow
        workflow_name: str = workflow.name
        # last_workflow = workflow_name

        logger.info("Running workflow: %s...", workflow_name)

        # workflow_start_time = time.time()
        # result = await workflow.run(context, callbacks)
        #
        # output = await emit_workflow_output(workflow)
        # yield PipelineRunResult(workflow_name, output, None)


def load_workflows(workflows_to_load):
    workflow_graph: dict[str, WorkflowToRun] = {}

    global anonymous_workflow_count
    for reference in workflows_to_load:
        name = reference.name
        is_anonymous = name is None or name.strip() == ""
        if is_anonymous:
            name = f"Anonymous Workflow {anonymous_workflow_count}"
            anonymous_workflow_count += 1
        name = cast(str, name)

        config = reference.config
        workflow = create_workflow(
            name or "MISSING NAME!",
            reference.steps,
            config,
        )
        workflow_graph[name] = WorkflowToRun(workflow, config=config or {})


def create_workflow(
    name: str,
    steps: list[PipelineWorkflowStep] | None = None,
    config: PipelineWorkflowConfig | None = None,
    memory_profile: bool = False,
):
    """Create a workflow from the given config."""

    steps = steps
    return Workflow(
        verbs={},
        schema={
            "name": name,
            "steps": steps,
        },
        validate=False,
        memory_profile=memory_profile,
    )



"""
`Representation Learning with Contrastive Predictive Coding
<https://arxiv.org/abs/1807.03748v2>`_
"""

import logging
from typing import Callable, Literal, Optional, Union

import torch
import torch.distributed.nn
import torch.nn as nn
import torch.nn.functional as F

logger = logging.getLogger(__name__)


class InfoNCE(nn.Module):
    """
    https://github.com/RElbers/info-nce-pytorch
    """

    def __init__(
        self,
        criterion: Union[nn.Module, Callable, None] = nn.CrossEntropyLoss(label_smoothing=0.0, reduction='mean'),
        temperature: float = 0.05,
        use_inbatch_negative: bool = True,
        negative_mode: Literal['paired', 'unpaired'] = "unpaired",
        **kwargs
    ):
        """
        if not normalized: temperature = 1.0, reset temperature = 1.0 due to using inner product to compute similarity
        if normalized: temperature should be smaller than 1.0 when use cosine similarity. Recommend to set it 0.01-0.1
        """
        super().__init__()
        self.criterion = criterion
        self.temperature = temperature
        self.use_inbatch_negative = use_inbatch_negative
        self.negative_mode = negative_mode
        if self.temperature > 0.5:
            logger.error('InfoNCE loss use normalized and inner product by default, temperature should be 0.01 ~ 0.1')

    def forward(
        self,
        query_embeddings: torch.Tensor,
        positive_embeddings: torch.Tensor,
        negative_embeddings: Optional[torch.Tensor] = None,
    ):
        query_embeddings = F.normalize(query_embeddings, dim=-1)
        positive_embeddings = F.normalize(positive_embeddings, dim=-1)
        device = query_embeddings.device
        if negative_embeddings is None:
            if self.negative_mode == 'unpaired':
                logits = query_embeddings @ positive_embeddings.transpose(-2, -1)
                target = torch.arange(logits.size(0), dtype=torch.long, device=device)
                loss = self.criterion(logits / self.temperature, target)
            else:
                logits1 = query_embeddings @ positive_embeddings.transpose(-2, -1)
                logits2 = logits1.T
                target = torch.arange(logits1.size(0), dtype=torch.long, device=device)
                loss = (
                    self.criterion(logits1 / self.temperature, target)
                    + self.criterion(logits2 / self.temperature, target)
                ) / 2
            return loss
        else:
            negative_embeddings = F.normalize(negative_embeddings, dim=-1)

            if self.use_inbatch_negative:
                logits = torch.cat([positive_embeddings, negative_embeddings], dim=0)
                similarity = query_embeddings @ logits.transpose(-2, -1)
                similarity = similarity / self.temperature
                similarity = similarity.view(query_embeddings.size(0), -1)
                target = torch.arange(query_embeddings.size(0), dtype=torch.long, device=device)
            else:
                # -> [batch_size, embedding_size, num_negative_samples]
                negative_embeddings = negative_embeddings.view(query_embeddings.size(0), -1, query_embeddings.size(1))
                negative_embeddings = negative_embeddings.permute(0, 2, 1)
                similarity = query_embeddings.unsqueeze(1) @ positive_embeddings.unsqueeze(2)
                negative_similarity = query_embeddings.unsqueeze(1) @ negative_embeddings
                similarity = torch.cat([similarity.squeeze(1), negative_similarity.squeeze(1)], dim=1)
                similarity = similarity / self.temperature
                target = torch.zeros(query_embeddings.size(0), dtype=torch.long, device=device)

            return self.criterion(similarity, target)



import torch
import torch.nn as nn


class BPRLoss(nn.Module):
    """Bayesian Personalized Ranking"""

    def __init__(self):
        super(BPRLoss, self).__init__()

    def forward(
        self,
        query_embeddings: torch.Tensor,
        positive_embeddings: torch.Tensor,
    ):
        return -(query_embeddings - positive_embeddings).sigmoid().log().sum()



"""
`Focal Loss for Dense Object Detection
<https://arxiv.org/abs/1708.02002>`_
"""

import torch
from torch import nn


class FocalLoss(nn.Module):
    def __init__(self, gamma: float = 0, eps: float = 1e-7):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.eps = eps
        self.ce = torch.nn.CrossEntropyLoss(reduction="none")

    def forward(self, inputs: torch.Tensor, labels: torch.Tensor):
        logp = self.ce(inputs, labels)
        p = torch.exp(-logp)
        loss = (1 - p) ** self.gamma * logp
        return loss.mean()



from typing import Dict, Iterable, List, Optional

import torch
from torch import nn


class MarginMSELoss(nn.Module):
    def __init__(self, scale: float = 1.0, **kwargs):
        super(MarginMSELoss, self).__init__()
        self.scale = scale
        self.criterion = nn.MSELoss()

    def forward(
        self,
        query_embeddings: torch.Tensor,
        positive_embeddings: torch.Tensor,
        negative_embeddings: Optional[torch.Tensor] = None,
        labels=None,
    ):
        scores_pos = (query_embeddings * positive_embeddings).sum(dim=-1) * self.scale
        scores_neg = (query_embeddings * negative_embeddings).sum(dim=-1) * self.scale
        margin_pred = scores_pos - scores_neg

        return self.criterion(margin_pred, labels)



import copy
import logging
import os
import time
from contextlib import nullcontext
from typing import TYPE_CHECKING, Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm, trange
from transformers import (
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    BatchEncoding,
    PreTrainedTokenizer,
)

from .base import Base
from .pooling import AutoPooling
from .utils import (
    batch_to_device,
    check_causal_lm,
    find_all_linear_names,
    get_device_name,
)

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)


class AutoModelForEmbedding(Base):
    """
    Loads or creates an Embedding model that can be used to map sentences / text.

    :param model_name_or_path: If it is a filepath on disc, it loads the model from that path. If it is not a path,
        it first tries to download a pre-trained SentenceTransformer model. If that fails, tries to construct a model
        from the Hugging Face Hub with that name.
    """

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        pooling_method: str = 'cls',
        normalize_embeddings: bool = False,
        max_length: Optional[int] = None,
        loss_fn: Optional[Callable] = None,
        query_instruction: Optional[str] = None,
        document_instruction: Optional[str] = None,
        use_fp16: bool = False,
        device: Optional[str] = None,
        **kwargs,
    ):
        """
        Loads or creates an Embedding model that can be used to map sentences / text.
        """
        super().__init__()
        if isinstance(model, str):
            assert ValueError("Please use AutoModelForEmbedding.from_pretrained(model_name_or_path)")
        self.model = model
        self.tokenizer = tokenizer
        self.pooling_method = pooling_method
        self.pooling = AutoPooling(pooling_method) if pooling_method else None
        self.loss_fn = loss_fn

        if max_length is None:
            if (
                hasattr(self.model, "config")
                and hasattr(self.model.config, "max_position_embeddings")
                and hasattr(self.tokenizer, "model_max_length")
            ):
                max_length = min(self.model.config.max_position_embeddings, self.tokenizer.model_max_length)
        else:
            logger.info('max_length will only work if the encode or forward function input text directly')

        self.max_length = max_length
        self.normalize_embeddings = normalize_embeddings

        self.query_instruction = query_instruction if query_instruction else ''
        self.document_instruction = document_instruction if document_instruction else ''
        self.use_fp16 = use_fp16
        self.device = device or get_device_name()
        self.model.to(self.device)

    def _init_weights(self, module: nn.Module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
        self,
        inputs,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = False,
    ):
        if isinstance(inputs, (dict, BatchEncoding)):
            embeddings = self.forward_from_loader(inputs['input_ids'], inputs['attention_mask'])
        elif isinstance(inputs, str) or (isinstance(inputs, list) and isinstance(inputs[0], str)):
            embeddings = self.forward_from_text(inputs)
        else:
            raise ValueError

        if labels is None or self.loss_fn is None:
            if return_dict:
                return {"sentence_embedding": embeddings}
            return embeddings
        else:
            outputs = dict()
            loss_output = self.loss_fn(embeddings, labels)
            outputs["loss"] = loss_output["loss"]
            outputs["sentence_embedding"] = loss_output["sentence_embedding"]
            return outputs

    def forward_from_loader(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, without_pooling: bool = False):
        model_output = self.model(input_ids, attention_mask=attention_mask, return_dict=True)
        if self.pooling is not None and not without_pooling:
            if 'last_hidden_state' in model_output:
                last_hidden_state = model_output['last_hidden_state']
            elif 'hidden_states' not in model_output:
                last_hidden_state = model_output[0]
            else:
                hidden_states = model_output['hidden_states']
                last_hidden_state = hidden_states[-1]
            embeddings = self.pooling(last_hidden_state, attention_mask=attention_mask)

            if self.normalize_embeddings:
                embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)

            return embeddings
        return model_output

    def forward_from_text(self, texts):
        batch_dict = self.tokenizer(
            texts,
            max_length=self.max_length,
            return_attention_mask=False,
            padding=False,
            truncation=True,
        )
        batch_dict["input_ids"] = [input_ids + [self.tokenizer.eos_token_id] for input_ids in batch_dict["input_ids"]]
        batch_dict = self.tokenizer.pad(batch_dict, padding=True, return_attention_mask=True, return_tensors="pt")
        batch_dict.pop("token_type_ids")
        return self.forward_from_loader(**batch_dict)

    def encode(
        self,
        inputs: Union[DataLoader, Dict, List, str],
        is_query: bool = False,
        batch_size: int = 16,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
    ):
        if isinstance(inputs, (DataLoader, BatchEncoding, Dict)):
            return self._encode_from_loader(
                loader=inputs,
                show_progress_bar=show_progress_bar,
                output_value=output_value,
                convert_to_numpy=convert_to_numpy,
                convert_to_tensor=convert_to_tensor,
                device=device,
                normalize_embeddings=normalize_embeddings,
            )
        elif isinstance(inputs, (str, List, Tuple, 'pd.Series', np.ndarray)):
            return self.encode_from_text(
                sentences=inputs,
                is_query=is_query,
                batch_size=batch_size,
                show_progress_bar=show_progress_bar,
                output_value=output_value,
                convert_to_numpy=convert_to_numpy,
                convert_to_tensor=convert_to_tensor,
                device=device,
                normalize_embeddings=normalize_embeddings,
            )
        else:
            raise ValueError(f'Input type: {type(inputs)}')

    def _encode_from_loader(
        self,
        loader: DataLoader,
        convert_to_numpy: bool = True,
        device: str = None,
        normalize_embeddings: bool = False,
        show_progress_bar: bool = None,
        **kwargs,
    ) -> Union[List[torch.Tensor], np.ndarray, torch.Tensor]:
        """Encode for sentence embedding"""
        device = device or self.device
        self.model.eval()
        self.model.to(device)

        all_embeddings = []

        for idx, inputs in enumerate(tqdm(loader, desc="Encoding", disable=not show_progress_bar)):
            with torch.autocast(device_type=device) if self.use_fp16 else nullcontext():
                with torch.no_grad():
                    inputs_on_device = {k: v.to(device) for k, v in inputs.items()}
                    embeddings = self.forward_from_loader(
                        inputs_on_device['input_ids'], attention_mask=inputs_on_device['attention_mask']
                    )
                    embeddings = embeddings.detach()
                    if normalize_embeddings:
                        embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)

                    # to avoid oom problems on gpu with large datasets
                    if convert_to_numpy:
                        embeddings = embeddings.cpu()
                    all_embeddings.append(embeddings)
        if convert_to_numpy:
            all_embeddings = np.concatenate([emb.numpy() for emb in all_embeddings], axis=0)
        else:
            all_embeddings = torch.concat(all_embeddings, dim=0)
        return all_embeddings

    def encode_from_text(
        self,
        sentences: Union[str, List[str], Tuple[str], 'pd.Series', np.ndarray],
        is_query: bool = False,
        batch_size: int = 16,
        show_progress_bar: bool = None,
        output_value: Literal["sentence_embedding", "token_embeddings", None] = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
    ) -> Union[List[torch.Tensor], np.ndarray, torch.Tensor]:
        """
        Computes sentence embeddings from sentence-transformers library.

        :param sentences: the sentences to embed.
        :param is_query: if the text is query or document
        :param batch_size: the batch size used for the computation.
        :param show_progress_bar: Whether to output a progress bar when encode sentences.
        :param output_value: The type of embeddings to return: "sentence_embedding" to get sentence embeddings,
            "token_embeddings" to get wordpiece token embeddings, and `None`, to get all output values. Defaults
            to "sentence_embedding".
        :param convert_to_numpy: Whether the output should be a list of numpy vectors. If False, it is a list of tensors
        :param convert_to_tensor: Whether the output should be one large tensor. Overwrites `convert_to_numpy`.
        :param device: Which `torch.device` to use for the computation.
        :param normalize_embeddings: Whether to normalize returned vectors to have length 1. In that case,
            the faster dot-product (util.dot_score) instead of cosine similarity can be used.

        :return: By default, a list of tensors is returned. If convert_to_tensor, a stacked tensor is returned.
            If convert_to_numpy, a numpy matrix is returned.
        """
        device = device or self.device
        self.model.eval()
        self.model.to(device)

        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG
            )

        if convert_to_tensor:
            convert_to_numpy = False

        if output_value != "sentence_embedding":
            convert_to_tensor = False
            convert_to_numpy = False

        input_was_string = False
        if isinstance(sentences, str) or not hasattr(sentences, "__len__"):
            sentences = [sentences]
            input_was_string = True

        all_embeddings = []
        length_sorted_idx = np.argsort([-self._text_length(sentence) for sentence in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]
        if is_query:
            logger.info("Encoding query")
            sentences_sorted = [self.query_instruction + sentence for sentence in sentences_sorted]
        else:
            logger.info('Encoding document')
            sentences_sorted = [self.document_instruction + sentence for sentence in sentences_sorted]

        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index : start_index + batch_size]
            features = self.tokenizer(
                sentences_batch,
                padding=True,
                truncation=True,
                return_tensors="pt",
                max_length=self.max_length,
            )
            features = batch_to_device(features, device)

            with torch.no_grad():
                out_features = self.forward(features, return_dict=True)

                if output_value == "token_embeddings":
                    embeddings = []
                    for token_emb, attention in zip(out_features[output_value], out_features["attention_mask"]):
                        last_mask_id = len(attention) - 1
                        while last_mask_id > 0 and attention[last_mask_id].item() == 0:
                            last_mask_id = last_mask_id - 1

                        embeddings.append(token_emb[0 : last_mask_id + 1])
                elif output_value is None:  # Return all outputs
                    embeddings = []
                    for sent_idx in range(len(out_features["sentence_embedding"])):
                        row = {name: out_features[name][sent_idx] for name in out_features}
                        embeddings.append(row)
                else:  # Sentence embeddings
                    embeddings = out_features[output_value]
                    embeddings = embeddings.detach()
                    if normalize_embeddings:
                        embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)

                    if convert_to_numpy:
                        embeddings = embeddings.cpu()

                all_embeddings.extend(embeddings)

        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]

        if convert_to_tensor:
            if len(all_embeddings):
                all_embeddings = torch.stack(all_embeddings)
            else:
                all_embeddings = torch.Tensor()
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_was_string:
            all_embeddings = all_embeddings[0]

        return all_embeddings

    def encode_queries(self, queries: List[str], **kwargs) -> np.ndarray:
        return self.encode_from_text(queries, is_query=True, **kwargs)

    def build_index(
        self,
        inputs: Union[DataLoader, Dict, List, str],
        index_path: Optional[str] = None,
        max_document_length: int = 512,
        split_documents: bool = False,
        batch_size: int = 16,
        show_progress_bar: bool = None,
        use_gpu: bool = False,
        save_id: bool = False,
    ):
        import faiss

        logger.info("Start to build index")
        start_time = time.time()

        embeddings = self.encode(
            inputs, batch_size=batch_size, convert_to_numpy=True, show_progress_bar=show_progress_bar
        )
        embeddings = np.asarray(embeddings, dtype=np.float32)

        index = faiss.IndexFlatL2(len(embeddings[0]))
        if save_id:
            # id_index = faiss.IndexFlatL2(len(embeddings[0]))
            index = faiss.IndexIDMap2(index)
            ids = inputs['ids']

        if use_gpu and self.device == 'cuda':
            logger.info('Build index use faiss-gpu')
            co = faiss.GpuMultipleClonerOptions()
            co.shard = True
            co.useFloat16 = True
            index = faiss.index_cpu_to_all_gpus(index, co=co)

        if save_id:
            index.add_with_ids(embeddings, ids)
        else:
            index.add(embeddings)

        if index_path:
            logger.info(f'Save faiss index to: {index_path}')
            if os.path.isdir(index_path):
                index_path = index_path + 'faiss.index'
            if not os.path.exists(os.path.dirname(index_path)):
                os.makedirs(os.path.dirname(index_path))

            if use_gpu and self.device == 'cuda':
                index = faiss.index_gpu_to_cpu(index)
            faiss.write_index(index, index_path)

        logger.info(f'Build index successfully, saved in {index_path}, elapsed: {time.time() - start_time:.2}s')
        return index

    def set_train_type(self, train_type: Literal['pointwise', 'pairwise', 'listwise'], **kwargs):
        train_type = train_type.lower().replace('-', '')
        logger.info(f'Set train type to {train_type}')
        model_class = {'pointwise': self, 'pairwise': PairwiseModel, 'listwise': ListwiseModel}
        model_class = model_class.get(train_type)

        return model_class(
            model=self.model,
            tokenizer=self.tokenizer,
            pooling_method=self.pooling_method,
            normalize_embeddings=self.normalize_embeddings,
            # loss_fn=self.loss_fn,
            query_instruction=self.query_instruction,
            document_instruction=self.document_instruction,
            device=self.device,
            **kwargs,
        )

    @classmethod
    def as_retriever(cls, retrieval_args, **kwargs):
        from .retrieval_auto import AutoModelForRetrieval

        embedding_model = cls(**kwargs)
        return AutoModelForRetrieval(embedding_model, **retrieval_args)

    @classmethod
    def as_reranker(cls, rerank_args, **kwargs):
        from .rerank import AutoModelForRanking

        ranking_model = cls(**kwargs)
        return AutoModelForRanking(ranking_model, **rerank_args)

    @classmethod
    def as_langchain_embedding(cls):
        from ..tools.langchain import LangchainEmbedding

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: Optional[str] = None,
        pooling_method: Optional[str] = "cls",
        pretrained: bool = True,
        config_path: Optional[str] = None,
        trust_remote_code: bool = True,
        custom_config_dict: Optional[Dict] = None,
        use_fp16: bool = False,
        use_lora: bool = False,
        lora_path: Optional[str] = None,
        lora_config=None,
        quantization_config=None,
        device: Optional[str] = None,
        query_instruction: Optional[str] = None,
        document_instruction: Optional[str] = None,
        max_length: Optional[int] = None,
        **kwargs,
    ):
        if not model_name_or_path or not isinstance(model_name_or_path, str):
            assert ValueError('Please input valid model_name_or_path')

        config = None
        if config_path:
            config = AutoConfig.from_pretrained(
                config_path, output_hidden_states=True, trust_remote_code=trust_remote_code
            )

        if custom_config_dict:
            if not config:
                config = AutoConfig.from_pretrained(
                    model_name_or_path, output_hidden_states=True, trust_remote_code=trust_remote_code
                )
            config.update(custom_config_dict)

        if check_causal_lm(model_name_or_path) and pooling_method != 'last':
            logger.warning('You are using a LLM model, while pooling_method is not last, is that sure?')

        if pretrained:
            model = AutoModel.from_pretrained(
                model_name_or_path,
                config=config,
                trust_remote_code=trust_remote_code,
                quantization_config=quantization_config,
                **kwargs,
            )
        else:
            model = AutoModel.from_config(config)
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code)

        if use_fp16:
            logger.info('Set model to fp16, please note that if you want fp16 during training, set training_args fp16')
            model.half()

        if use_lora and lora_path is None:
            logger.info('Set fine-tuning to LoRA')
            from peft import LoraConfig, TaskType, get_peft_model

            if lora_config is None:
                lora_alpha = 64
                lora_dropout = 0.05
                target_modules = find_all_linear_names(model)
                lora_config = LoraConfig(
                    lora_alpha=lora_alpha,
                    lora_dropout=lora_dropout,
                    target_modules=target_modules,
                    bias='none',
                    task_type='FEATURE_EXTRACTION',
                )

            model = get_peft_model(model, lora_config)
            model.print_trainable_parameters()

        if lora_path is not None:
            logger.info(f'Load pretrained with LoRA adapter {lora_path}')
            from peft import LoraConfig, PeftModel

            model = PeftModel.from_pretrained(model, lora_path)
            model = model.merge_and_unload()

        return cls(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            device=device,
            query_instruction=query_instruction,
            document_instruction=document_instruction,
            use_fp16=use_fp16,
            max_length=max_length,
            **kwargs,
        )


class PairwiseModel(AutoModelForEmbedding):
    """Pairwise Model wrapper
    - bi_encoder
        - shared_weights or not
    - cross_encoder
    - poly_encoder
    support: query + pos pair, or query + pos + neg triplet
    """

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        pooling_method: str = 'cls',
        normalize_embeddings: bool = False,
        loss_fn: Optional[Callable] = None,
        cross_encoder: bool = False,
        poly_encoder: bool = False,
        shared_weights: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            normalize_embeddings=normalize_embeddings,
            loss_fn=loss_fn,
            **kwargs,
        )

        self.cross_encoder = cross_encoder
        self.shared_weights = shared_weights
        if not shared_weights:
            self.document_model = copy.deepcopy(self.model)

    def forward(
        self,
        inputs: Union[Dict[str, torch.Tensor], list],
        inputs_pair: Optional[Dict[str, torch.Tensor]] = None,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        if isinstance(inputs, (list, tuple, dict)) and 2 <= len(inputs) <= 3 or inputs_pair is not None:
            if inputs_pair:
                input1 = inputs
                input2 = inputs_pair
            elif isinstance(inputs, dict):
                inputs = list(inputs.values())
            if isinstance(inputs, (list, tuple)):
                input1 = inputs[0]
                input2 = inputs[1]
                if len(inputs) == 3:
                    input3 = inputs[2]

            ids1, mask1 = input1['input_ids'], input1['attention_mask']
            ids2, mask2 = input2['input_ids'], input2['attention_mask']
            if self.cross_encoder:
                ids = torch.cat([ids1, ids2], dim=0)
                mask = torch.cat([mask1, mask2], dim=0)

                transformer_out = super().forward_from_loader(
                    {"input_ids": ids, "attention_mask": mask}, without_pooling=True
                )
                pooled_output = self.pooling(transformer_out[0], mask)
                pooled_output1 = pooled_output[: len(ids1), :]
                pooled_output2 = pooled_output[len(ids1) :, :]

            else:
                # bi-encoder, pooling in each
                if self.shared_weights:
                    pooled_output1 = super().forward_from_loader(ids1, attention_mask=mask1)
                    pooled_output2 = super().forward_from_loader(ids2, attention_mask=mask2)
                    if len(inputs) == 3:
                        pooled_output3 = super().forward_from_loader(
                            input3['input_ids'], attention_mask=input3['attention_mask']
                        )
                        if self.loss_fn is None:
                            return pooled_output1, pooled_output2, pooled_output3
                        outputs = dict()
                        loss = self.loss_fn(pooled_output1, pooled_output2, pooled_output3)
                        outputs["loss"] = loss
                        return outputs

                else:
                    pooled_output1 = super().forward_from_loader(ids1, attention_mask=mask1)
                    pooled_output2 = self.document_model(ids2, mask2)

            if self.loss_fn is None:
                return pooled_output1, pooled_output2
            else:
                outputs = dict()
                loss = self.loss_fn(pooled_output1, pooled_output2)
                outputs["loss"] = loss
                return outputs

        else:
            pooled_output = super(PairwiseModel, self).forward_from_loader(**inputs, without_pooling=False)
            return pooled_output


class ListwiseModel(AutoModelForEmbedding):
    """
    Listwise model by segment_id
    """

    def __init__(
        self,
        model: Optional[nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        pooling_method: str = 'cls',
        normalize_embeddings: bool = False,
        loss_fn: Optional[Callable] = None,
        listwise_pooling: bool = False,
        num_segments: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            model=model,
            tokenizer=tokenizer,
            pooling_method=pooling_method,
            normalize_embeddings=normalize_embeddings,
            loss_fn=loss_fn,
            **kwargs,
        )
        self.pooling_method = pooling_method
        self.listwise_pooling = listwise_pooling
        self.num_segments = num_segments

    def forward(
        self,
        inputs,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        encoding = super().forward(inputs)

        res = dict()
        if self.pooling_method == 'unsorted_segment_mean':
            encoding = self._unsorted_segment_mean(
                encoding, segment_ids=inputs['segment_ids'], num_segments=self.num_segments
            )
            res['pred'] = self.fc(encoding[:, 1:])
        else:
            encodings = []
            for i in range(self.num_segments):
                mask_ = (inputs['segment_ids'] == i + 1).int()
                encoding_ = self.pooling(encoding, mask_)
                encoding_ = self.fc(encoding_)
                encodings.append(encoding_)
            res['pred'] = torch.stack(encodings, 1)

        res['pred'] = res['pred'].squeeze(-1)
        return res

    def _unsorted_segment_mean(self, data: torch.Tensor, segment_ids: torch.Tensor, num_segments: int) -> torch.Tensor:
        result_shape = (num_segments, data.size(1))
        segment_ids = segment_ids.unsqueeze(-1).expand(-1, data.size(1))  # (batch, num_embedding)
        result = data.new_full(result_shape, 0)  # init empty result tensor
        count = data.new_full(result_shape, 0)
        result.scatter_add_(0, segment_ids, data)  # fill the result from data to organized segment result
        count.scatter_add_(0, segment_ids, torch.ones_like(data))
        return result / count.clamp(min=1)

    def _sorted_segment_mean(self, data: torch.Tensor, segment_ids: torch.Tensor, num_segments: int) -> torch.Tensor:
        """
        Compute the mean of each segment in data based on sorted segment_ids.

        Args:
            data (torch.Tensor): Input data tensor of shape (batch_size, num_embedding).
            segment_ids (torch.Tensor): Sorted segment IDs tensor of shape (batch_size,).
            num_segments (int): Number of unique segments.

        Returns:
            torch.Tensor: Tensor of shape (num_segments, num_embedding) containing the mean of each segment.
        """
        result = torch.zeros((num_segments, data.size(1)), dtype=data.dtype, device=data.device)
        count = torch.zeros((num_segments,), dtype=torch.int32, device=data.device)

        start_idx = 0
        for i in range(num_segments):
            """Find the range of indices corresponding to the current segment"""
            while start_idx < segment_ids.size(0) and segment_ids[start_idx] == i:
                start_idx += 1

            if start_idx > 0 and segment_ids[start_idx - 1] == i:
                segment_slice = slice(start_idx - (start_idx - segment_ids[start_idx:].tolist().count(i)), start_idx)
                result[i] = data[segment_slice].sum(dim=0)
                count[i] = segment_slice.stop - segment_slice.start

        result /= count.clamp(min=1).unsqueeze(-1)
        return result



VERSION = (0, 0, 11)

__version__ = ".".join(map(str, VERSION))
short_version = __version__


import sys

msg = "Open-retrievals is only compatible with Python 3.7 and newer, please consider a newer version."

if sys.version_info < (3, 6):
    raise ImportError(msg)



from .collator import (
    ColBertCollator,
    EncodeCollator,
    LLMRerankCollator,
    PairCollator,
    RerankCollator,
    TripletCollator,
)
from .dataset import EncodeDataset, RerankTrainDataset, RetrievalTrainDataset
from .sampler import GroupedBatchSampler, GroupSortedBatchSampler



"""
`R-Drop: Regularized Dropout for Neural Networks
<https://arxiv.org/abs/2106.14448>`_
"""

import torch
from torch import nn


class RDropLoss(nn.Module):
    def __init__(self, alpha: float = 4, reduction: str = "mean", **kwargs):
        super(RDropLoss, self).__init__()
        self.reduction = reduction
        self.alpha = alpha
        self.entropy = nn.CrossEntropyLoss(reduction=reduction)
        self.kl = nn.KLDivLoss(reduction=reduction)

    def forward(self, inputs: torch.Tensor, labels: torch.Tensor):
        loss = self.entropy(inputs, labels.long())
        loss_kl = self.kl(inputs[::2], inputs[1::2]) + self.kl(inputs[1::2], inputs[::2])
        return loss + loss_kl / 4 * self.alpha



import re
from typing import Dict, List, Literal, Optional

import torch
from torch import nn
from transformers import PreTrainedModel

DEFAULT_LLM_PATTERNS = [r'.*llama.*', r'.*mistral.*', r'.*qwen.*', r'.*baichuan.*', r'.*intern.*', r'.*Phi.*']


def get_device_name() -> Literal["mps", "cuda", "cpu"]:
    """
    Returns the name of the device where this module is running on.
    It's a simple implementation that doesn't cover cases when more powerful GPUs are available and
    not a primary device ('cuda:0') or MPS device is available, but not configured properly:
    https://pytorch.org/docs/master/notes/mps.html

    :return: Device name, like 'cuda' or 'cpu'
    """
    if torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    else:
        return "cpu"


def batch_to_device(batch: Dict, target_device: str) -> Dict[str, torch.Tensor]:
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], torch.Tensor):
            batch[key] = batch[key].to(target_device)
        else:
            batch[key] = torch.tensor(batch[key], dtype=torch.long).to(target_device)
    return batch


def check_causal_lm(model_name_or_path: str, llm_regex_patterns: List[str] = None) -> bool:
    """check if it's a decoder-only causal model"""
    if llm_regex_patterns is not None:
        llm_regex_patterns += DEFAULT_LLM_PATTERNS
    else:
        llm_regex_patterns = DEFAULT_LLM_PATTERNS
    model_name_or_path = model_name_or_path.lower()
    for pattern in llm_regex_patterns:
        if re.match(pattern, model_name_or_path):
            return True
    return False


def find_all_linear_names(model: PreTrainedModel, linear_type: Optional[object] = None) -> List[str]:
    """
    Find all linear layer names

    :param model: PreTrainedModel
    :param linear_type: Optional[object] = None, linear type, such as nn.Linear, bnb.nn.Linear4bit, bnb.nn.Linear8bitLt

    :return: List[str], linear layer names
    """
    if linear_type is None:
        linear_type = nn.Linear
    lora_module_names = set()
    for name, module in model.named_modules():
        if isinstance(module, linear_type):
            names = name.split('.')
            lora_module_names.add(names[0] if len(names) == 1 else names[-1])

    if 'lm_head' in lora_module_names:  # needed for 16-bit
        lora_module_names.remove('lm_head')
    return list(lora_module_names)


def resize_token_embeddings(
    model,
    new_num_tokens: Optional[int] = None,
    pad_to_multiple_of: Optional[int] = None,
) -> nn.Embedding:
    """when you modify the tokenizer, such as adding new tokens or changing the vocabulary size"""
    return model.resize_token_embeddings(new_num_tokens=new_num_tokens, pad_to_multiple_of=pad_to_multiple_of)


def save_swa_weights(model: nn.Module, model_path_list: List[str], save_file: str, device: str):
    """Get the swa weights from a list of model weights"""

    def average_state_dicts(state_dicts: List[dict]) -> dict:
        """Average the state dictionaries."""
        averaged_state = {}
        num_states = len(state_dicts)
        for key in state_dicts[0]:
            averaged_state[key] = sum(state_dict[key] for state_dict in state_dicts) / num_states
        return averaged_state

    state_list = [torch.load(path, map_location=device) for path in model_path_list]
    averaged_state = average_state_dicts(state_list)
    msg = model.load_state_dict(averaged_state, strict=False)
    print(f"State dict load message: {msg}")

    model.half()
    torch.save(model.state_dict(), save_file)
    print(f"Checkpoint saved at: {save_file}")


def freeze_layers(model, n_layers: int = 6):
    """Freeze layers before the last n_layers."""
    trainable_layers = 0
    for name, param in model.named_parameters():
        if param.requires_grad:
            trainable_layers += 1

    for index, (name, param) in enumerate(iterable=model.named_parameters()):
        if index < (trainable_layers - n_layers):
            param.requires_grad = False

    return model



"""
file parser for RAG
https://github.com/InternLM/HuixiangDou
"""

import hashlib
import io
import logging
from typing import Dict, List

import pandas as pd
import requests

logger = logging.getLogger(__name__)


class BaseParser:
    def parse(self, file_path) -> List[Dict]:
        raise NotImplementedError()


class PdfParser(BaseParser):
    def __init__(self, parser: str):
        self.parser = parser


def process_pdf_plumber(file_path):
    import pdfplumber

    with pdfplumber.open(file_path) as pdf:
        text = ''
        for i, page in enumerate(pdf.pages):
            text = page.extract_text()
            if text:
                lines = text.split('\n')
                text_wo_header_footer = '\n'.join(lines[1:-1])
                text += text_wo_header_footer + '\n\n'

            if page.extract_tables():
                for table in page.extract_tables():
                    markdown_table = ''
                    for i, row in enumerate(table):
                        row = [cell for cell in row if cell]
                        processed_row = [str(cell).strip() if cell is not None else '' for cell in row]
                        markdown_row = '| ' + '| '.join(processed_row) + '|\n'
                        markdown_table += markdown_row
                        if i == 0:
                            sep = [':----' if cell.isdigit() else '----' for cell in row]
                            markdown_table += '| ' + '| '.join(sep) + '|\n'
                    text += markdown_table + '\n'
    return text


def process_pdf(file_path):
    import pypdf

    text = ""
    with open(file_path, 'rb') as file:
        reader = pypdf.PdfReader(file)
        for page in reader.pages:
            text += page.extract_text()
    return text


def process_pdf2(file_path):
    import PyPDF2

    text = ""
    with open(file_path, 'rb') as file:
        reader = PyPDF2.PdfReader(file)
        for page in reader.pages:
            text += page.extract_text().strip()
    return text


def process_url(url):
    response = requests.get(url)
    if url.endswith('.pdf'):
        return process_pdf(io.BytesIO(response.content))
    elif url.endswith('.txt'):
        return response.text
    else:
        return "Unsupported file format"


def process_txt(file_path, text_splitter):
    with open(file_path, 'r', encoding='utf-8') as file:
        text = file.read()
    chunks = text_splitter.create_documents([text])
    documents = []
    for chunk in chunks:
        chunk.metadata = {'source': file_path, 'read': file_path}
        documents.append(chunk)
    return documents


def process_word(file_path):
    from docx import Document

    doc = Document(file_path)
    text = ''
    for paragraph in doc.paragraphs:
        text_ = paragraph.text.strip()
        if text_:
            text += text_
            text += '\n'
    return text


def process_excel(file_path: str):
    if file_path.endswith('.csv'):
        table = pd.read_csv(file_path)
    else:
        table = pd.read_excel(file_path)
    if table is None:
        return ''
    json_text = table.dropna(axis=1).to_json(force_ascii=False)
    return json_text


def process_markdown(file_path: str, head_splitter, md_splitter, text_splitter):
    text = ''
    with open(file_path, encoding='utf8') as f:
        text = f.read()

    docs = head_splitter.split_text(text)
    final = []
    for doc in docs:
        header = ''
        if len(doc.metadata) > 0:
            if 'Header 1' in doc.metadata:
                header += doc.metadata['Header 1']
            if 'Header 2' in doc.metadata:
                header += ' '
                header += doc.metadata['Header 2']
            if 'Header 3' in doc.metadata:
                header += ' '
                header += doc.metadata['Header 3']

        if len(doc.page_content) >= 1024:
            subdocs = md_splitter.create_documents([doc.page_content])
            for subdoc in subdocs:
                if len(subdoc.page_content) >= 10:
                    final.append('{} {}'.format(header, subdoc.page_content.lower()))
        elif len(doc.page_content) >= 10:
            final.append('{} {}'.format(header, doc.page_content.lower()))  # noqa E501

        return final


def process_epub(file_path: str):
    import ebooklib
    from bs4 import BeautifulSoup

    book = ebooklib.epub.read_epub(file_path)

    text_content = []
    for item in book.get_items():
        if item.get_type() == ebooklib.ITEM_DOCUMENT:
            text_content.append(item.get_content().decode("utf-8"))

    wt = "\n".join(text_content)
    sp = BeautifulSoup(wt, "html.parser")
    text = sp.get_text()
    return text


class FileParser:
    """File parser class RAG upload file"""

    def __init__(self):
        self.image_suffix = ['.jpg', '.jpeg', '.png', '.bmp']
        self.md_suffix = ['.md']
        self.text_suffix = ['.txt', '.text']
        self.excel_suffix = ['.xlsx', '.xls', '.csv']
        self.pdf_suffix = ['.pdf']
        self.ppt_suffix = ['.pptx']
        self.html_suffix = ['.html', '.htm', '.shtml', '.xhtml']
        self.word_suffix = ['.docx', '.doc']
        self.normal_suffix = (
            self.md_suffix
            + self.text_suffix
            + self.excel_suffix
            + self.pdf_suffix
            + self.word_suffix
            + self.ppt_suffix
            + self.html_suffix
        )

    def read(self, file_path: str, to_document: bool = False):
        file_type = self.get_type(file_path)
        text = ''

        try:
            if file_type == 'md' or file_type == 'text':
                with open(file_path) as f:
                    text = f.read()

            elif file_type == 'pdf':
                text += process_pdf(file_path)

            elif file_type == 'excel':
                text += process_excel(file_path)

            elif file_type == 'word' or file_type == 'ppt':
                import textract

                text = textract.process(file_path).decode('utf8')
                if file_type == 'ppt':
                    text = text.replace('\n', ' ')

            elif file_type == 'html':
                from bs4 import BeautifulSoup

                with open(file_path) as f:
                    soup = BeautifulSoup(f.read(), 'html.parser')
                    text += soup.text

        except Exception as e:
            logger.error((file_path, str(e)))
            return '', e

        text = text.replace('\n\n\n', '\n')
        text = text.replace('\n\n', '\n')
        text = text.replace('  ', ' ')
        return text, None

    def get_type(self, file_path: str):
        for suffix in self.pdf_suffix:
            if file_path.endswith(suffix):
                return 'pdf'

        for suffix in self.md_suffix:
            if file_path.endswith(suffix):
                return 'md'

        for suffix in self.ppt_suffix:
            if file_path.endswith(suffix):
                return 'ppt'

        for suffix in self.image_suffix:
            if file_path.endswith(suffix):
                return 'image'

        for suffix in self.text_suffix:
            if file_path.endswith(suffix):
                return 'text'

        for suffix in self.word_suffix:
            if file_path.endswith(suffix):
                return 'word'

        for suffix in self.excel_suffix:
            if file_path.endswith(suffix):
                return 'excel'

        for suffix in self.html_suffix:
            if file_path.endswith(suffix):
                return 'html'
        return None

    def md5(self, file_path: str):
        hash_object = hashlib.sha256()
        with open(file_path, 'rb') as file:
            chunk_size = 8192
            while chunk := file.read(chunk_size):
                hash_object.update(chunk)

        return hash_object.hexdigest()[0:8]



import torch
import torch.nn as nn


class DiceLoss(nn.Module):
    def __init__(self, smooth=1.0, reduction='mean'):
        super(DiceLoss, self).__init__()
        self.smooth = smooth
        self.reduction = reduction

    def forward(self, pred, target):
        # Ensure contiguous tensors
        pred = pred.contiguous()
        target = target.contiguous()

        intersection = (pred * target).sum(dim=2).sum(dim=2)

        loss = 1 - (
            (2.0 * intersection + self.smooth)
            / (pred.sum(dim=2).sum(dim=2) + target.sum(dim=2).sum(dim=2) + self.smooth)
        )

        # Apply reduction method
        if self.reduction == 'mean':
            loss = loss.mean()
        elif self.reduction == 'sum':
            loss = loss.sum()

        return loss


class Dice(nn.Module):
    def __init__(self, feature_num):
        super(Dice, self).__init__()
        self.alpha = nn.Parameter(torch.ones(1))
        self.feature_num = feature_num
        self.bn = nn.BatchNorm1d(self.feature_num, affine=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x_norm = self.bn(x)
        x_p = self.sigmoid(x_norm)
        return self.alpha * (1.0 - x_p) * x + x_p * x



"""Reranking fine tune pipeline"""

import logging
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

import torch
import transformers
from torch import nn
from transformers import (
    AdamW,
    AutoTokenizer,
    HfArgumentParser,
    TrainingArguments,
    get_linear_schedule_with_warmup,
    set_seed,
)

from ..data import (
    ColBertCollator,
    RerankCollator,
    RerankTrainDataset,
    RetrievalTrainDataset,
)
from ..data.collator import LLMRerankCollator
from ..losses import ColbertLoss, TokenLoss
from ..models.rerank import AutoModelForRanking, ColBERT
from ..trainer import RerankTrainer

# os.environ["WANDB_LOG_MODEL"] = "false"
transformers.logging.set_verbosity_error()
logger = logging.getLogger(__name__)


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """

    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None, metadata={"help": "Where do you want to store the pretrained models downloaded from s3"}
    )
    causal_lm: bool = field(default=False, metadata={'help': "Whether the model is a causal lm or not"})
    colbert_dim: int = field(default=1024)


@dataclass
class DataArguments:
    data_name_or_path: str = field(default=None, metadata={"help": "Path to corpus"})
    train_group_size: int = field(default=8)
    unfold_each_positive: bool = field(default=False)
    query_max_length: int = field(default=64)
    max_length: int = field(
        default=256,
        metadata={
            "help": "The maximum total input sequence length after tokenization for input text. Sequences longer "
            "than this will be truncated, sequences shorter will be padded."
        },
    )
    query_key: str = field(default=None)
    positive_key: str = field(default=None)
    negative_key: str = field(default=None)
    query_instruction: str = field(default=None, metadata={"help": "instruction for query"})
    document_instruction: str = field(default=None, metadata={"help": "instruction for document"})
    task_prompt: str = field(
        default=(
            "Given a query A and a passage B, determine whether the passage contains an answer "
            "to the query by providing a prediction of either 'Yes' or 'No'."
        )
    )

    def __post_init__(self):
        self.dataset_split = 'train'
        self.dataset_language = 'default'

        if self.data_name_or_path is not None:
            if not os.path.isfile(self.data_name_or_path) and not os.path.isdir(self.data_name_or_path):
                info = self.data_name_or_path.split('/')
                self.dataset_split = info[-1] if len(info) == 3 else 'train'
                self.data_name_or_path = "/".join(info[:-1]) if len(info) == 3 else '/'.join(info)
                self.dataset_language = 'default'
                if ':' in self.data_name_or_path:
                    self.data_name_or_path, self.dataset_language = self.data_name_or_path.split(':')


@dataclass
class RerankerTrainingArguments(TrainingArguments):
    model_type: str = field(default='cross-encoder', metadata={'help': "train type of cross-encoder, colbert"})
    negatives_cross_device: bool = field(default=False, metadata={"help": "share negatives across devices"})
    use_inbatch_negative: bool = field(default=False)
    temperature: Optional[float] = field(default=0.02)
    remove_unused_columns: bool = field(default=False)
    num_train_epochs: int = field(default=3)
    use_lora: bool = field(default=False)
    use_bnb_config: bool = field(default=False)
    do_rerank: bool = field(default=False, metadata={"help": "run the reranking loop"})
    report_to: Optional[List[str]] = field(
        default="none", metadata={"help": "The list of integrations to report the results and logs to."}
    )


def get_optimizer(model, learning_rate, weight_decay=0.0):
    optimizer_parameters = [
        {
            "params": [p for n, p in model.model.named_parameters()],
            "lr": learning_rate,
            "weight_decay": weight_decay,
        },
        {
            "params": [p for n, p in model.named_parameters() if "model" not in n],
            "lr": learning_rate * 20,
            "weight_decay": 0.0,
        },
    ]
    return AdamW(optimizer_parameters)


def main():
    parser = HfArgumentParser((ModelArguments, DataArguments, RerankerTrainingArguments))
    model_args, data_args, training_args = parser.parse_args_into_dataclasses()
    model_args: ModelArguments
    data_args: DataArguments
    training_args: TrainingArguments

    if (
        os.path.exists(training_args.output_dir)
        and os.listdir(training_args.output_dir)
        and training_args.do_train
        and not training_args.overwrite_output_dir
    ):
        raise ValueError(
            f"Output directory ({training_args.output_dir}) is not empty. Use --overwrite_output_dir to overcome."
        )

    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO if training_args.local_rank in [-1, 0] else logging.WARN,
    )
    logger.warning(
        "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
        training_args.local_rank,
        training_args.device,
        training_args.n_gpu,
        bool(training_args.local_rank != -1),
        training_args.fp16,
    )
    logger.info("Training/evaluation parameters %s", training_args)
    logger.info("Model parameters %s", model_args)
    logger.info("Data parameters %s", data_args)

    set_seed(training_args.seed)

    tokenizer = AutoTokenizer.from_pretrained(
        model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
        cache_dir=model_args.cache_dir,
        use_fast=False,
    )

    if training_args.use_bnb_config:
        from transformers import BitsAndBytesConfig

        logger.info('Use quantization bnb config')
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_use_double_quant=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_compute_dtype=torch.bfloat16,
        )
    else:
        quantization_config = None

    if training_args.model_type == 'colbert':
        logger.info('Set rank model to ColBERT')
        train_dataset = RetrievalTrainDataset(
            args=data_args,
            tokenizer=tokenizer,
            train_group_size=data_args.train_group_size,
            unfold_each_positive=data_args.unfold_each_positive,
            positive_key=data_args.positive_key,
            negative_key=data_args.negative_key,
            query_instruction=data_args.query_instruction,
            document_instruction=data_args.document_instruction,
        )
        data_collator = ColBertCollator(
            tokenizer,
            query_max_length=data_args.query_max_length,
            document_max_length=data_args.max_length,
            positive_key=data_args.positive_key,
            negative_key=data_args.negative_key,
        )
        model = ColBERT.from_pretrained(
            model_args.model_name_or_path,
            colbert_dim=model_args.colbert_dim,
            loss_fn=ColbertLoss(temperature=0.02, use_inbatch_negative=training_args.use_inbatch_negative),
        )
    elif training_args.model_type == 'cross-encoder':
        logger.info('Set rank model to Cross-Encoder')
        train_dataset = RerankTrainDataset(args=data_args, tokenizer=tokenizer)
        data_collator = RerankCollator(tokenizer, max_length=data_args.max_length)
        model = AutoModelForRanking.from_pretrained(
            model_args.model_name_or_path,
            num_labels=1,
            loss_fn=nn.BCEWithLogitsLoss(reduction='mean'),
            causal_lm=model_args.causal_lm,
        )
    elif training_args.model_type == 'llm':
        logger.info('Set rank model to LLM')
        train_dataset = RetrievalTrainDataset(
            args=data_args,
            tokenizer=tokenizer,
            unfold_each_positive=data_args.unfold_each_positive,
            train_group_size=data_args.train_group_size,
            positive_key=data_args.positive_key,
            negative_key=data_args.negative_key,
        )
        data_collator = LLMRerankCollator(
            tokenizer=tokenizer, max_length=data_args.max_length, prompt=data_args.task_prompt
        )
        token_index = tokenizer('Yes', add_special_tokens=False)['input_ids'][-1]
        model = AutoModelForRanking.from_pretrained(
            model_args.model_name_or_path,
            num_labels=1,
            loss_fn=TokenLoss(token_index=token_index, train_group_size=data_args.train_group_size),
            causal_lm=True,
            use_lora=training_args.use_lora,
            quantization_config=quantization_config,
        )
    else:
        raise ValueError(
            f"model_type should be one of 'colbert', 'cross-encoder' and 'llm', instead of {training_args.model_type}"
        )

    logger.info(f"Total training examples: {len(train_dataset)}")
    optimizer = get_optimizer(model, learning_rate=training_args.learning_rate)

    num_train_steps = int(
        len(train_dataset) / training_args.per_device_train_batch_size * training_args.num_train_epochs
    )
    scheduler = get_linear_schedule_with_warmup(
        optimizer, num_warmup_steps=0.05 * num_train_steps, num_training_steps=num_train_steps
    )

    trainer = RerankTrainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        data_collator=data_collator,
        tokenizer=tokenizer,
    )
    trainer.optimizer = optimizer
    trainer.scheduler = scheduler

    Path(training_args.output_dir).mkdir(parents=True, exist_ok=True)

    trainer.train()
    # trainer.save_model(training_args.output_dir)
    model.save_pretrained(training_args.output_dir)

    if trainer.is_world_process_zero():
        tokenizer.save_pretrained(training_args.output_dir)


if __name__ == "__main__":
    main()



"""A simple implementation for workflow and pipeline refer to Microsoft GraphRAG"""

import gc
import logging
import time
from dataclasses import dataclass as dc_dataclass
from dataclasses import field
from typing import Any, Generic, List, Optional, TypeVar, cast

from pydantic import BaseModel
from pydantic import Field as pydantic_Field

"""Represent a step in a workflow."""
PipelineWorkflowStep = dict[str, Any]

"""Represent a configuration for a workflow."""
PipelineWorkflowConfig = dict[str, Any]


class PipelineWorkFlowReference(BaseModel):
    """Represent a reference to a workflow, and can optionally be the workflow itself."""

    name: Optional[str] = pydantic_Field(description="Name of the workflow.", default=None)
    steps: List[PipelineWorkflowStep] = pydantic_Field(description="The optional steps for the workflow.", default=None)
    config: PipelineWorkflowConfig = pydantic_Field(
        description="The optional configuration for the workflow.", default=None
    )


@dc_dataclass
class PipelineRunStats:
    """Pipeline running stats."""

    total_runtime: float = field(default=0)
    """Float representing the total runtime."""

    num_documents: int = field(default=0)
    """Number of documents."""

    input_load_time: float = field(default=0)
    """Float representing the input load time."""

    workflows: dict[str, dict[str, float]] = field(default_factory=dict)
    """A dictionary of workflows."""


Context = TypeVar("Context")


class Workflow(Generic[Context]):
    # https://github.com/microsoft/datashaper/blob/main/python/datashaper/datashaper/workflow/workflow.py
    def __init__(self):
        pass


@dc_dataclass
class WorkflowToRun:
    """Workflow to run class definition."""

    workflow: 'workflow'
    config: dict[str, Any]



from torch import nn

from .arcface import ArcFaceAdaptiveMarginLoss
from .circle import MultiLabelCircleLoss
from .colbert_loss import ColbertLoss
from .cosent import CoSentLoss
from .cosine_similarity import CosineSimilarity
from .infonce import InfoNCE
from .pearson_loss import PearsonLoss
from .simcse import SimCSE
from .token_loss import TokenLoss
from .triplet import TripletLoss


class AutoLoss(nn.Module):
    def __init__(self, loss_name, loss_kwargs):
        super(AutoLoss, self).__init__()
        if loss_name == 'simcse':
            self.loss_fn = SimCSE(**loss_kwargs)
        elif loss_name == 'triplet':
            self.loss_fn = TripletLoss(**loss_kwargs)
        elif loss_name == 'infonce':
            self.loss_fn = InfoNCE(**loss_kwargs)
        elif loss_name == 'arcface':
            self.loss_fn = ArcFaceAdaptiveMarginLoss(**loss_kwargs)
        elif loss_name == 'cosent':
            self.loss_fn = CoSentLoss(**loss_kwargs)
        elif loss_name == 'pearson':
            self.loss_fn = PearsonLoss(**loss_kwargs)

    def forward(self, *args, **kwargs):
        return self.loss_fn(*args, **kwargs)



from typing import Dict, List


def get_recall(qid2positive: Dict[str, List[str]], qid2ranking: Dict[str, List[str]], cutoff_ranks: List[int] = [10]):
    qid2recall = {cutoff_rank: {} for cutoff_rank in cutoff_ranks}
    num_samples = len(qid2ranking.keys())

    for qid in qid2positive:
        positives = qid2positive[qid]
        ranked_doc_ids = qid2ranking[qid]

        for rank, doc_id in enumerate(ranked_doc_ids, start=1):
            if doc_id in positives:
                for cutoff_rank in cutoff_ranks:
                    if rank <= cutoff_rank:
                        qid2recall[cutoff_rank][qid] = qid2recall[cutoff_rank].get(qid, 0) + 1.0 / len(positives)

    return {
        f"recall@{cutoff_rank}": sum(qid2recall[cutoff_rank].values()) / num_samples for cutoff_rank in cutoff_ranks
    }



from typing import Callable, Optional, Union

import torch
from torch import nn
from torch.nn import functional as F


class NTXcent(nn.Module):
    def __init__(
        self,
        criterion: Union[nn.Module, Callable] = nn.CrossEntropyLoss(label_smoothing=0.0, reduction='mean'),
        temperature=0.05,
        **kwargs
    ):
        super().__init__()
        self.criterion = criterion
        self.temperature = temperature

    def forward(self, x):
        similarity = F.cosine_similarity(x[None, :, :], x[:, None, :], dim=-1)
        similarity[torch.eye(x.size(0)).bool()] = float("-inf")
        similarity = similarity / self.temperature

        target = torch.arange(x.size(0))
        target[0::2] += 1
        target[1::2] -= 1

        return F.cross_entropy(similarity, target, reduction="mean")



"""
Token loss for LLM generative token compared with the expected token
"""

from typing import Callable, Literal, Optional, Union

import torch
import torch.distributed.nn
import torch.nn as nn


class TokenLoss(nn.Module):
    def __init__(
        self,
        token_index: int,
        train_group_size: int = 2,
        criterion: Union[nn.Module, Callable, None] = nn.CrossEntropyLoss(label_smoothing=0.0, reduction='mean'),
        **kwargs
    ):
        super().__init__()
        self.token_index = token_index
        self.train_group_size = train_group_size
        self.criterion = criterion

    def forward(self, logits: torch.Tensor, labels: torch.Tensor):
        batch_size = labels.size(0) // self.train_group_size
        _, max_indices = torch.max(labels, dim=1)
        # shift the targets such that output n predicts token n+1
        predict_indices = max_indices - 1
        logits = [logits[i, predict_indices[i], :] for i in range(logits.shape[0])]
        logits = torch.stack(logits, dim=0)
        scores = logits[:, self.token_index]

        grouped_scores = scores.contiguous().view(batch_size, -1)
        target = torch.zeros(batch_size, device=grouped_scores.device, dtype=torch.long)
        loss = self.criterion(grouped_scores, target)
        return loss



import math
from typing import Dict, List


def get_ndcg(qid2positive: Dict[str, List[str]], qid2ranking: Dict[str, List[str]], cutoff_rank: int = 10):
    """
    qid2positive: {qid: [pos1_doc_id, pos2_doc_id]}
    qid2ranking: {qid: [rank1_doc_id, rank2_doc_id, rank3_doc_id]}
    """

    def dcg(relevance_scores: List[int], cutoff: int) -> float:
        return sum([rel / math.log2(idx + 2) for idx, rel in enumerate(relevance_scores[:cutoff])])

    qid2ndcg = dict()

    for qid in qid2positive:
        positives = set(qid2positive[qid])
        ranked_doc_ids = qid2ranking[qid]

        relevance_scores = [1 if doc_id in positives else 0 for doc_id in ranked_doc_ids]
        actual_dcg = dcg(relevance_scores, cutoff_rank)

        ideal_relevance_scores = sorted(relevance_scores, reverse=True)
        ideal_dcg = dcg(ideal_relevance_scores, cutoff_rank)

        if ideal_dcg > 0:
            qid2ndcg[qid] = actual_dcg / ideal_dcg
        else:
            qid2ndcg[qid] = 0.0

    return {f"ndcg@{cutoff_rank}": sum(qid2ndcg.values()) / len(qid2ranking.keys())}



"""
`ArcFace: Additive Angular Margin Loss for Deep Face Recognition
<https://arxiv.org/abs/1801.07698>`_
"""

import logging
import math
from typing import Callable, Optional, Union

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

logger = logging.getLogger(__name__)


class ArcFaceAdaptiveMarginLoss(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        criterion: Union[nn.Module, Callable, None] = None,
        scale: float = 30.0,
        margin: float = 0.50,
        easy_margin: bool = False,
        eps: float = 0.0,
        **kwargs
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.criterion = criterion
        self.scale = scale

        self.eps = eps
        self.arc_weight = nn.Parameter(torch.FloatTensor(out_features, in_features))
        self.init_parameters()

        self.easy_margin = easy_margin
        self.set_margin(margin)

    def init_parameters(self) -> None:
        nn.init.xavier_uniform_(self.arc_weight)

    def set_margin(self, margin: float):
        self.margin = margin
        self.cos_m = torch.from_numpy(np.asarray(np.cos(margin))).float()
        self.sin_m = torch.from_numpy(np.asarray(np.sin(margin))).float()
        self.th = nn.Parameter(torch.FloatTensor([math.cos(math.pi - margin)]), requires_grad=False)
        self.mm = nn.Parameter(
            torch.FloatTensor([math.sin(math.pi - margin) * margin]),
            requires_grad=False,
        )

    def forward(self, embeddings: torch.Tensor, labels: torch.Tensor, weight=None, margin: Optional[float] = None):
        if margin is not None:
            # dynamic margin
            self.set_margin(margin=margin)

        cosine = F.linear(F.normalize(embeddings), F.normalize(self.arc_weight))
        sine = torch.sqrt(1.0 - torch.pow(cosine, 2))
        phi = cosine * self.cos_m - sine * self.sin_m

        if self.easy_margin:
            phi = torch.where(cosine > 0, phi, cosine)
        else:
            phi = torch.where(cosine > self.th, phi, cosine - self.mm)

        if len(labels.shape) > 1 and labels.shape[1] > 1:
            one_hot = labels
        else:
            device = embeddings.device
            one_hot = torch.zeros(cosine.size(), device=device)
            one_hot.scatter_(1, labels.view(-1, 1).long(), 1)

        if self.eps > 0:
            one_hot = (1 - self.eps) * one_hot + self.eps / self.out_features

        output = (one_hot * phi) + ((1.0 - one_hot) * cosine)
        output *= self.scale

        output_dict = dict()
        output_dict["sentence_embedding"] = output
        if self.criterion:
            output_dict["loss"] = self.criterion(output, one_hot, weight)
        return output_dict



"""LLM fine tune pipeline"""

import copy
import logging
import os
from dataclasses import dataclass, field
from functools import partial
from typing import Dict, List, Optional

import transformers
from datasets import load_dataset
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    DataCollatorForSeq2Seq,
    Trainer,
)

from ..models.utils import find_all_linear_names

logger = logging.getLogger(__name__)

PROMPT = (
    "Below is an instruction that describes a task. "
    "Write a response that appropriately completes the request.\n\n"
    "### Instruction:\n{instruction}\n\n### Response:"
)
IGNORE_INDEX = -100


@dataclass
class ModelArguments:
    model_name_or_path: Optional[str] = field(default=None)


@dataclass
class DataArguments:
    data_path: str = field(
        default=None,
        metadata={"help": "Path to the training data."},
    )
    source_length: int = field(default=512)
    target_length: int = field(default=512)


@dataclass
class TrainingArguments(transformers.TrainingArguments):
    output_dir: Optional[str] = field(default=None)
    cache_dir: Optional[str] = field(default=None)
    num_train_epochs = (1,)
    per_device_train_batch_size: int = field(default=1)
    gradient_accumulation_steps: int = field(default=4)
    optim: str = field(default="paged_adamw_32bit")  # "adamw_torch"
    model_max_length: int = field(
        default=512,
        metadata={"help": "Maximum sequence length. Sequences will be right padded (and possibly truncated)."},
    )
    use_lora: bool = field(default=True)
    save_steps: int = field(default=100)
    logging_steps: int = field(default=50)
    learning_rate: float = field(default=2e-4)
    max_grad_norm: float = field(default=0.3)
    # max_steps: int = field(default=1000)
    warmup_ratio: float = field(default=0.03)
    lr_scheduler_type: str = field(default="constant")
    remove_unused_columns: bool = field(default=False)
    group_by_length: bool = field(
        default=False,
        metadata={
            "help": "Group sequences into batches with same length. Saves memory and speeds up training considerably."
        },
    )
    use_deepspeed: bool = field(default=False)


def build_sft_data(data_args):
    datasets = load_dataset(
        "json",
        data_files={"train": data_args.data_path},
        # cache_dir=cache_dir,
    )
    train_dataset = datasets["train"]
    return train_dataset


def build_prompt(examples, prompt_input, tokenizer, data_args):
    if "instruction" in examples:
        ins_data = examples["instruction"]
        input_data = examples["input"]
    else:
        ins_data = examples["input"]
        input_data = [""] * len(ins_data)
    output = examples["output"]

    len_ = len(ins_data)
    sources = [
        (
            prompt_input.format_map({"instruction": ins_data[i], "input": input_data[i]})
            if input_data[i] != ""
            else prompt_input.format_map({"instruction": ins_data[i]})
        )
        for i in range(len_)
    ]
    sources = [i[: data_args.source_length] for i in sources]
    targets = [f"{example[:data_args.target_length-1]}{tokenizer.eos_token}" for example in output]
    return sources, targets


def tokenize_fn(strings, tokenizer, IGNORE_INDEX):
    """Tokenize a list of strings."""
    tokenized_list = [
        tokenizer(
            text,
            return_tensors="pt",
            padding="longest",
            max_length=tokenizer.model_max_length,
            truncation=True,
        )
        for text in strings
    ]
    input_ids = labels = [tokenized.input_ids[0] for tokenized in tokenized_list]
    ne_pad_token_id = IGNORE_INDEX if tokenizer.pad_token_id is None else tokenizer.pad_token_id
    input_ids_lens = labels_lens = [
        tokenized.input_ids.ne(ne_pad_token_id).sum().item() for tokenized in tokenized_list
    ]
    return dict(
        input_ids=input_ids,
        labels=labels,
        input_ids_lens=input_ids_lens,
        labels_lens=labels_lens,
    )


def preprocess(examples, tokenizer, data_args):
    sources, targets = build_prompt(examples, prompt_input=PROMPT, tokenizer=tokenizer, data_args=data_args)

    examples = [s + t for s, t in zip(sources, targets)]
    examples_tokenized, sources_tokenized = [
        tokenize_fn(strings, tokenizer, IGNORE_INDEX) for strings in (examples, sources)
    ]
    input_ids = examples_tokenized["input_ids"]
    labels = copy.deepcopy(input_ids)
    for label, source_len in zip(labels, sources_tokenized["input_ids_lens"]):
        label[:source_len] = IGNORE_INDEX
    return dict(input_ids=input_ids, labels=labels)


def build_model(
    model_args: ModelArguments,
    training_args: TrainingArguments,
) -> tuple:
    model = AutoModelForCausalLM.from_pretrained(
        model_args.model_name_or_path,
        cache_dir=training_args.cache_dir,
        torch_dtype="auto",
        # if model_args.model_name_or_path.find("falcon") != -1 else False
        trust_remote_code=True,
    )

    if training_args.use_lora:
        from peft import LoraConfig, get_peft_model

        lora_alpha = 128
        LORA_DROPOUT = 0.05
        TARGET_MODULES = find_all_linear_names(model)

        config = LoraConfig(
            lora_alpha=lora_alpha,
            target_modules=TARGET_MODULES,
            lora_dropout=LORA_DROPOUT,
            bias="none",
            task_type="CAUSAL_LM",
        )
        # model = model.to(torch.bfloat16)
        model = get_peft_model(model, config)
        model.print_trainable_parameters()

    # model.is_parallelizable = True
    # model.model_parallel = True
    # torch.cuda.empty_cache()
    return model


def train():
    parser = transformers.HfArgumentParser((ModelArguments, DataArguments, TrainingArguments))
    model_args, data_args, training_args = parser.parse_args_into_dataclasses()
    print(training_args)

    tokenizer = AutoTokenizer.from_pretrained(
        model_args.model_name_or_path,
        # use_fast=False,
        encode_special_tokens=True,
        trust_remote_code=True,
        # model_max_length=training_args.model_max_length,
        cache_dir=training_args.cache_dir,
    )
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.padding_side = "right"

    train_dataset = build_sft_data(data_args)

    train_dataset = train_dataset.map(
        function=partial(preprocess, tokenizer=tokenizer, data_args=data_args),
        batched=True,
        remove_columns=train_dataset.column_names,
        desc="Running tokenizer on train dataset",
        num_proc=20,
    ).shuffle()

    model = build_model(model_args, training_args)
    data_collator = DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model, label_pad_token_id=IGNORE_INDEX)

    checkpoint = None
    if training_args.resume_from_checkpoint is not None:
        checkpoint = training_args.resume_from_checkpoint

    trainer = Trainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=None,
        data_collator=data_collator,
    )
    trainer.train(resume_from_checkpoint=checkpoint)
    trainer.save_state()
    trainer.save_model(output_dir=training_args.output_dir)


if __name__ == "__main__":
    train()



from .data.collator import (
    ColBertCollator,
    LLMRerankCollator,
    PairCollator,
    RerankCollator,
    TripletCollator,
)
from .data.dataset import RerankTrainDataset, RetrievalTrainDataset
from .models.embedding_auto import AutoModelForEmbedding, ListwiseModel, PairwiseModel
from .models.pooling import AutoPooling
from .models.rerank import AutoModelForRanking, ColBERT
from .models.retrieval_auto import AutoModelForRetrieval
from .trainer.custom_trainer import CustomTrainer
from .trainer.trainer import RerankTrainer, RetrievalTrainer
from .trainer.tuner import AutoTuner



"""RAG pipeline"""

import argparse
import logging
import re
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, TypeVar, Union

from ..tools.file_parser import FileParser
from ..tools.generator import BaseLLM
from ..tools.prompts import RAG_PROMPT
from ..tools.router import Router

logger = logging.getLogger(__name__)
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
    level=logging.INFO,
)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--tokenizer_name', required=False)
    return parser.parse_args()


class RAGConfig(object):
    def __init__(self, top_n: int = 3):
        self.top_n = top_n

    @classmethod
    def from_dict(
        cls,
    ):
        return RAGConfig()


class SimpleRAG(object):
    def __init__(self, embedder, retriever, generator, reranker=None):
        self.embedder = embedder
        self.retriever = retriever
        self.reranker = reranker
        self.generator = generator

    def build(self):
        pass

    def add(self):
        pass

    def load(self):
        pass

    def search(self):
        pass

    def generate(self, query: str, context: Optional[str] = None):
        """LLM response"""
        prompt_kwargs = {'context': context, 'query': query}
        response = self.generator(prompt_kwargs)
        return response

    @classmethod
    def from_dict(cls):
        """Create an instance from previously serialized data"""
        obj = cls.__new__(cls)
        return obj


def index_process():
    return


def retrieval_process(query, top_k: int = 3):
    return


def rerank_process():
    return


def chat_process(llm, prompt, history):
    response, history = llm.chat(prompt, history)
    return response, history


def rag_process(prompt, corpus_path, top_k: int = 3):
    global llm, history

    context = retrieval_process(prompt, top_k=top_k)

    context = rerank_process(prompt, context)

    prompt_with_context = RAG_PROMPT.format(question=prompt, context="\n".join(context))

    response, history = chat_process(prompt_with_context, history)
    return


class Session(object):
    def __init__(self, query: str, history: list):
        self.query = query
        self.history = history


def extract_citations(bullet):
    # matches digits or commas
    matches = re.findall(r"\[([\d, ]+)\]", bullet)
    ref_ids = []
    for match in matches:
        ref_ids += [int(m.strip()) for m in match.split(",") if len(m.strip()) > 0]
    return ref_ids


def main():
    args = parse_args()
    print(args)
    return


if __name__ == '__main__':
    main()



"""
https://arxiv.org/pdf/2404.16130v1
https://github.com/microsoft/graphrag

Graph, community, entity
"""

import logging

from ..models.retrieval_auto import GraphRetrieval
from ..tools.refiner import LLMRefiner

logger = logging.getLogger(__name__)



import logging
import math
import os
import random
from typing import Dict, Iterable, List, Optional, Tuple, Union

import datasets
from torch.utils.data import Dataset
from transformers import BatchEncoding, PreTrainedTokenizer

logger = logging.getLogger(__name__)


class RetrievalTrainDataset(Dataset):
    """Retrieval training dataset"""

    def __init__(
        self,
        data_name_or_path: Union[str, datasets.Dataset, None] = None,
        train_group_size: int = 2,
        unfold_each_positive: bool = False,
        query_key: str = 'query',
        positive_key: str = 'positive',
        negative_key: str = 'negative',
        query_instruction: str = '',
        document_instruction: str = '',
        separator: str = ' ',
        args: Optional = None,
        tokenizer: PreTrainedTokenizer = None,
        dataset_split: str = 'train',
        dataset_language: str = 'default',
    ):
        if args:
            data_name_or_path = args.data_name_or_path
            self.train_group_size = (
                args.train_group_size if 'train_group_size' in args.__dataclass_fields__ else train_group_size
            )
            self.dataset_split = args.dataset_split if 'dataset_split' in args.__dataclass_fields__ else dataset_split
            self.dataset_language = (
                args.dataset_language if 'dataset_language' in args.__dataclass_fields__ else 'default'
            )
            self.unfold_each_positive = (
                args.unfold_each_positive
                if 'unfold_each_positive' in args.__dataclass_fields__
                else unfold_each_positive
            )
            self.query_instruction = args.query_instruction if args.query_instruction else query_instruction
            if self.query_instruction is None:
                self.query_instruction = ''
            self.document_instruction = args.document_instruction if args.document_instruction else document_instruction
            if self.document_instruction is None:
                self.document_instruction = ''
            self.query_key = args.query_key or query_key
            self.positive_key = args.positive_key or positive_key
            self.negative_key = args.negative_key or negative_key

        else:
            self.train_group_size = train_group_size
            self.dataset_split = dataset_split
            self.dataset_language = dataset_language
            self.query_instruction = query_instruction
            self.document_instruction = document_instruction
            self.query_key = query_key
            self.positive_key = positive_key
            self.negative_key = negative_key
            self.unfold_each_positive = unfold_each_positive

        if isinstance(data_name_or_path, datasets.Dataset):
            dataset = data_name_or_path
        elif os.path.isdir(data_name_or_path):
            train_datasets = []
            for file in os.listdir(data_name_or_path):
                temp_dataset = datasets.load_dataset(
                    "json",
                    data_files=os.path.join(data_name_or_path, file),
                )
                train_datasets.append(temp_dataset)
            dataset = datasets.concatenate_datasets(train_datasets)
        else:
            if data_name_or_path.endswith('jsonl') or data_name_or_path.endswith('json'):
                dataset = datasets.load_dataset("json", data_files=data_name_or_path)
            else:
                dataset = datasets.load_dataset(data_name_or_path, self.dataset_language)

        if self.dataset_split in dataset:  # train or dev
            dataset = dataset[dataset_split]
        dataset = dataset.filter(lambda x: len(x[self.positive_key]) > 0)
        if self.negative_key in dataset[0]:
            dataset = dataset.filter(lambda x: len(x[self.negative_key]) > 0)

        self.unfold_each_positive = unfold_each_positive
        self.tokenizer = tokenizer
        self.args = args
        self.query_key = query_key
        self.positive_key = positive_key
        self.negative_key = negative_key
        self.separator = separator

        logger.info("Load original {} retrieval data.".format(len(dataset)))

        if self.unfold_each_positive:
            self.samples = self.generate_unfold_samples(dataset)
        else:
            self.dataset = dataset
        logger.info(
            "Generate total {} retrieval data. Query instruction: {}, Document instruction: {}".format(
                len(self.dataset), self.query_instruction, self.document_instruction
            )
        )

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, item: int) -> Union[Dict[str, str], List[BatchEncoding]]:
        if self.unfold_each_positive:
            return self.samples[item]

        data = self.dataset[item]
        query = data[self.query_key]
        if self.query_instruction:
            query = self.query_instruction + query

        if isinstance(data[self.positive_key], (list, tuple)):
            if isinstance(data[self.positive_key][0], dict):
                pos = random.choice(data[self.positive_key])
                pos_text = pos['title'] + self.separator + pos['text'] if 'title' in pos else pos['text']
                pos = self.document_instruction + pos_text
            else:
                pos = self.document_instruction + random.choice(data[self.positive_key])
        else:
            pos = self.document_instruction + data[self.positive_key]

        sample = {self.query_key: query, self.positive_key: pos}
        if self.negative_key in data:
            if isinstance(data[self.negative_key], (list, tuple)):
                if len(data[self.negative_key]) < self.train_group_size - 1:
                    num = math.ceil((self.train_group_size - 1) / len(data[self.negative_key]))
                    negs = random.sample(data[self.negative_key] * num, self.train_group_size - 1)
                else:
                    negs = random.sample(data[self.negative_key], self.train_group_size - 1)

            else:
                negs = [data[self.negative_key]]

            if isinstance(negs[0], dict):
                negs = [neg['title'] + self.separator + neg['text'] if 'title' in neg else neg['text'] for neg in negs]

            sample.update({self.negative_key: [self.document_instruction + neg for neg in negs]})
        return sample

    def generate_unfold_samples(self, dataset):
        samples: List = []
        for data in dataset:
            for pos_text in data[self.positive_key]:
                sample = {
                    self.query_key: self.query_instruction + data[self.query_key],
                    self.positive_key: self.document_instruction + pos_text,
                }

                if self.negative_key in data:
                    if isinstance(data[self.negative_key], (list, tuple)):
                        if len(data[self.negative_key]) < self.train_group_size - 1:
                            num = math.ceil((self.train_group_size - 1) / len(data[self.negative_key]))
                            negs = random.sample(data[self.negative_key] * num, self.train_group_size - 1)
                        else:
                            negs = random.sample(data[self.negative_key], self.train_group_size - 1)
                    else:
                        negs = data[self.negative_key]
                    sample.update({self.negative_key: [self.document_instruction + neg for neg in negs]})
                samples.append(sample)

        return samples

    def dynamic_sample(self, batch_size: int, missing_list=None, wrong_dict=None, max_wrong: int = 16):
        logger.info('Dynamic Shuffle Sample')
        return


class RerankTrainDataset(Dataset):
    """Rerank training dataset"""

    def __init__(
        self,
        data_name_or_path: Optional[str] = None,
        train_group_size: int = 2,
        unfold_each_positive: bool = False,
        query_key: str = 'query',
        positive_key: Optional[str] = 'document',
        negative_key: Optional[str] = 'negative',
        args: Optional = None,
        tokenizer: PreTrainedTokenizer = None,
        dataset_split: str = 'train',
    ):
        """
        train_group_size = 1(positive) + max_negative_samples
        """
        if args:
            data_name_or_path = args.data_name_or_path
            self.train_group_size = (
                args.train_group_size if 'train_group_size' in args.__dataclass_fields__ else train_group_size
            )
            self.dataset_split = args.dataset_split if 'dataset_split' in args.__dataclass_fields__ else dataset_split
            self.unfold_each_positive = (
                args.unfold_each_positive
                if 'unfold_each_positive' in args.__dataclass_fields__
                else unfold_each_positive
            )
            self.query_key = args.query_key or query_key
            self.positive_key = args.positive_key or positive_key
            self.negative_key = args.negative_key or negative_key
        else:
            self.train_group_size = train_group_size
            self.dataset_split = dataset_split
            self.query_key = query_key
            self.positive_key = positive_key
            self.negative_key = negative_key
            self.unfold_each_positive = unfold_each_positive

        if isinstance(data_name_or_path, datasets.Dataset):
            dataset = data_name_or_path
        elif os.path.isdir(data_name_or_path):
            train_datasets = []
            for file in os.listdir(data_name_or_path):
                temp_dataset = datasets.load_dataset(
                    "json",
                    data_files=os.path.join(data_name_or_path, file),
                )

                train_datasets.append(temp_dataset)
            dataset = datasets.concatenate_datasets(train_datasets)

        else:
            if data_name_or_path.endswith('jsonl') or data_name_or_path.endswith('json'):
                dataset = datasets.load_dataset("json", data_files=data_name_or_path)
            else:
                dataset = datasets.load_dataset(data_name_or_path)

        if dataset_split in dataset:
            dataset = dataset[dataset_split]
        dataset = dataset.filter(lambda x: len(x[self.positive_key]) > 0)
        if self.negative_key in dataset[0]:
            dataset = dataset.filter(lambda x: len(x[self.negative_key]) > 0)

        logger.info("Load original {} rerank data.".format(len(dataset)))
        if positive_key:
            dataset = self.generate_samples(dataset)

        self.dataset = dataset
        logger.info("Generate total {} rerank data.".format(len(self.dataset)))

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item: int):
        if isinstance(self.dataset[item], dict):
            query = self.dataset[item][self.query_key]
            document = self.dataset[item][self.positive_key]
            labels = self.dataset[item]['labels']
        else:
            query, document, labels = self.dataset[item]
        sample = {"query": query, "document": document, "labels": labels}
        return sample

    def generate_samples(self, dataset):
        max_negative_samples = None
        if self.train_group_size and self.train_group_size > 0:
            max_negative_samples = self.train_group_size - 1

        samples: List = []
        for data in dataset:
            if self.unfold_each_positive:
                for pos_text in data[self.positive_key]:
                    samples.append([data[self.query_key], pos_text, 1])
            else:
                samples.append([data[self.query_key], random.choice(data[self.positive_key]), 1])

            negative_samples = data[self.negative_key]
            if max_negative_samples:
                if len(negative_samples) < max_negative_samples:
                    num = math.ceil(max_negative_samples / len(negative_samples))
                    negative_samples = random.sample(negative_samples * num, max_negative_samples)
                else:
                    negative_samples = random.sample(negative_samples, max_negative_samples)
            for neg_text in negative_samples:
                samples.append([data[self.query_key], neg_text, 0])
        return samples


class EncodeDataset(Dataset):
    """Retrieval inference dataset"""

    def __init__(
        self,
        data_name_or_path: Union[str, datasets.Dataset, None] = None,
        max_length: int = 128,
        id_key: Optional[str] = None,
        text_key: str = 'query',
        instruction: str = '',
        args: Optional = None,
        tokenizer: PreTrainedTokenizer = None,
        dataset_split: str = 'train',
        dataset_language: str = 'default',
    ):
        if args:
            data_name_or_path = args.data_name_or_path
            dataset_language = args.dataset_language if 'dataset_language' in args.__dataclass_fields__ else 'default'
            dataset_split = args.dataset_split if 'dataset_split' in args.__dataclass_fields__ else 'train'
            text_key = args.query_key
            instruction = args.query_instruction or args.document_instruction or instruction

        if isinstance(data_name_or_path, datasets.Dataset):
            self.encode_data = data_name_or_path
        else:
            try:
                self.encode_data = datasets.load_dataset("json", data_files=data_name_or_path)
            except FileNotFoundError:
                self.encode_data = datasets.load_dataset(data_name_or_path, dataset_language)

            if dataset_split in self.encode_data:
                self.encode_data = self.encode_data[dataset_split]

        self.tokenizer = tokenizer
        self.max_length = max_length
        self.id_key = id_key
        self.text_key = text_key
        self.instruction = instruction
        if len(instruction) > 0:
            logger.info(f'Add prefix instruction: {self.instruction}')

    def __len__(self):
        return len(self.encode_data)

    def __getitem__(self, item) -> [str, BatchEncoding]:
        if self.id_key is not None:
            text_id, text = (self.encode_data[item][f] for f in [self.id_key, self.text_key])
            text = self.instruction + text
            encoded_text = self.tokenizer.encode_plus(
                text,
                max_length=self.max_length,
                truncation=True,
                padding=False,
                return_token_type_ids=False,
            )
            return text_id, encoded_text
        else:
            text = self.encode_data[item][self.text_key]
            text = self.instruction + text
            encoded_text = self.tokenizer.encode_plus(
                text,
                max_length=self.max_length,
                truncation='only_first',
                padding=False,
                return_token_type_ids=False,
            )
            return encoded_text


class RerankDataset(Dataset):
    """Rerank inference dataset"""

    def __init__(self):
        pass



"""Matryoshka Representation Learning"""

from typing import List, Optional

import torch
from torch import nn


class MRLLayer(nn.Module):
    def __init__(self, mrl_nested_dim: List[int]):
        super().__init__()
        # self.projection = nn.Linear()
        self.mrl_nested_dim = mrl_nested_dim

    def forward(self, x):
        # x = self.projection(x)
        nested_x = []
        for dim in self.mrl_nested_dim:
            nested_x.append(x[:, :dim])
        return nested_x


class MRLLoss(nn.Module):
    """
    https://arxiv.org/pdf/2205.13147
    """

    def __init__(self, criterion: nn.Module, mrl_nested_dim: List[int]):
        super().__init__()
        self.criterion = criterion
        self.mrl_nested_dim = mrl_nested_dim
        self.query_mrl = MRLLayer(self.mrl_nested_dim)
        self.positive_mrl = MRLLayer(self.mrl_nested_dim)
        self.negative_mrl = MRLLayer(self.mrl_nested_dim)

    def forward(
        self,
        query_embeddings: torch.Tensor,
        pos_embeddings: torch.Tensor,
        neg_embeddings: Optional[torch.Tensor] = None,
    ):
        query_mrl_embed = self.query_mrl(query_embeddings)
        positive_mrl_embed = self.positive_mrl(pos_embeddings)
        if neg_embeddings is not None:
            negative_mrl_embed = self.negative_mrl(neg_embeddings)
        else:
            negative_mrl_embed = [None] * len(positive_mrl_embed)

        loss = torch.tensor(0, device=query_embeddings.device)
        for query_embed, positive_embed, negative_embed in zip(query_mrl_embed, positive_mrl_embed, negative_mrl_embed):
            loss += self.criterion(query_embed, positive_embed, negative_embed)

        loss = loss / len(self.mrl_nested_dim)

        return loss



import torch
import torch.nn as nn
import torch.nn.functional as F


class CoSentLoss(nn.Module):
    def __init__(self, temperature: float = 0.05, **kwargs):
        super().__init__()
        self.temperature = temperature

    def forward(self, embed1: torch.Tensor, embed2: torch.Tensor, labels=None):
        """
        https://github.com/bojone/CoSENT/blob/124c368efc8a4b179469be99cb6e62e1f2949d39/cosent.py
        """
        pred_sims = F.cosine_similarity(embed1, embed2)
        # each position cosine of every pair
        pred_sims = pred_sims[:, None] - pred_sims[None, :]
        pred_sims = pred_sims - (1 - labels) * 1e12
        pred_sims = pred_sims.view(-1)
        # e^0 = 1, so it equals to add 1 in log
        pred_sims = torch.cat([torch.tensor([0.0], device=embed1.device), pred_sims], dim=0)
        return torch.logsumexp(pred_sims, dim=0)



"""Integration with LlamaIndex"""

from typing import Any, Dict, List, Optional

from llama_index.bridge.pydantic import Field, PrivateAttr
from llama_index.callbacks import CBEventType, EventPayload
from llama_index.core.embeddings import BaseEmbedding
from llama_index.core.retrievers import BaseRetriever
from llama_index.postprocessor.types import BaseNodePostprocessor
from llama_index.schema import MetadataMode, NodeWithScore, QueryBundle
from llama_index.utils import infer_torch_device

from ..models.embedding_auto import AutoModelForEmbedding
from ..models.rerank import AutoModelForRanking
from .generator import BaseLLM


class LlamaIndexEmbedding(AutoModelForEmbedding, BaseEmbedding):
    client: Any
    model_name_or_path: Optional[str] = None
    cache_folder: Optional[str] = None
    model_kwargs: Dict[str, Any] = dict()
    encode_kwargs: Dict[str, Any] = dict()

    def __init__(self, embed_dim: int, **kwargs: Any):
        BaseEmbedding.__init__(self)
        if 'model_name' in kwargs:
            kwargs['model_name_or_path'] = kwargs.pop('model_name')
        AutoModelForEmbedding.__init__(self, **kwargs)


class LlamaIndexRetrieval(BaseRetriever):
    def __init__(
        self,
        vector_store,
        embed_model: BaseEmbedding,
        similarity_top_k: int = 2,
    ):
        super(LlamaIndexRetrieval, self).__init__()


class LlamaIndexReranker(BaseNodePostprocessor):
    model: str = Field(ddescription="Sentence transformer model name.")
    top_n: int = Field(description="Number of nodes to return sorted by score.")
    _model: Any = PrivateAttr()

    def __init__(self, model: str, top_n: int = 5, device: Optional[str] = None, **kwargs):
        self.model = AutoModelForRanking.from_pretrained(model)
        device = infer_torch_device() if device is None else device
        super().__init__(model=model, top_n=top_n, device=device)

    @classmethod
    def class_name(cls):
        return 'OpenRetrievalsRank'

    def _postprocess_nodes(
        self,
        nodes: List[NodeWithScore],
        query_bundle: Optional[QueryBundle] = None,
    ):
        if query_bundle is None:
            raise ValueError("Missing query bundle in extra info.")
        if len(nodes) == 0:
            return []

        query = query_bundle.query_str
        passages = []
        valid_nodes = []
        invalid_nodes = []
        for node in nodes:
            passage = node.node.get_content(metadata_mode=MetadataMode.EMBED)
            if isinstance(passage, str) and len(passage) > 0:
                passages.append(passage.replace('\n', ' '))
                valid_nodes.append(node)
            else:
                invalid_nodes.append(node)

        with self.callback_manager.event(
            CBEventType.RERANKING,
            payload={
                EventPayload.NODES: nodes,
                EventPayload.MODEL_NAME: self.model,
                EventPayload.QUERY_STR: query_bundle.query_str,
                EventPayload.TOP_K: self.top_n,
            },
        ) as event:
            rerank_result = self._model.rerank(query, passages)
            new_nodes = []
            for score, nid in zip(rerank_result['rerank_scores'], rerank_result['rerank_ids']):
                node = valid_nodes[nid]
                node.score = score
                new_nodes.append(node)
            for node in invalid_nodes:
                node.score = 0
                new_nodes.append(node)

            assert len(new_nodes) == len(nodes)

            new_nodes = new_nodes[: self.top_n]
            event.on_end(payload={EventPayload.NODES: new_nodes})

        return new_nodes


class LlamaIndexLLM(BaseLLM):
    def __init__(self):
        pass

    def generate(self, prompt: str) -> str:
        return



"""LLM generator"""

from abc import ABC, abstractmethod
from typing import Generic, Protocol, TypeVar


class BaseLLM(ABC):
    """Base class for LLM chat"""

    @abstractmethod
    def generate(self, prompt: str, max_length: int) -> str:
        """Generate LLM response"""

    @abstractmethod
    async def agenerate(self, prompt: str, max_length: int):
        """Generate LLM response async"""


class BaseLLMCallback:
    """Base class for LLM callback"""

    def __init__(self):
        self.response = []



import gc
import logging
import math
import time
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
from accelerate import Accelerator
from torch import nn
from tqdm import tqdm

from .adversarial import AWP, EMA, FGM

logger = logging.getLogger(__name__)


def train_fn(
    epoch: int,
    model: nn.Module,
    train_loader,
    optimizer,
    criterion: Optional[Callable] = None,
    apex: bool = False,
    gradient_accumulation_steps: int = 1,
    max_grad_norm: float = 1,
    data_collator=None,
    batch_scheduler=None,
    fgm=None,
    awp=None,
    ema_inst=None,
    print_freq: int = 100,
    wandb: bool = False,
    device: str = "cuda",
    teacher: Optional[nn.Module] = None,
    teacher_loss: Optional[Callable] = None,
    **kwargs,
):
    start = time.time()
    global_step = 0
    model.train()
    scaler = torch.cuda.amp.GradScaler(enabled=apex)
    losses = AverageMeter()
    save_step = int(len(train_loader) / 1)

    # loss_list = []
    # metrics_list = []

    for step, inputs in enumerate(train_loader):
        if isinstance(inputs, (list, tuple)) and len(inputs) == 2:
            inputs, labels = inputs
        elif isinstance(inputs, dict):
            # assert data_collator is not None, "if labels are in inputs keys, can't set data_collator to None"
            labels = inputs['labels']
            inputs.pop('labels', None)

        if data_collator:
            inputs = data_collator(inputs)

        if isinstance(inputs, (list, tuple)):
            for input in inputs:
                if isinstance(input, dict):
                    for k, v in input.items():
                        input[k] = v.to(device)
                else:
                    input = input.to(device)
        else:
            for k, v in inputs.items():
                inputs[k] = v.to(device)
        labels = labels.to(device)

        batch_size = labels.size(0)
        with torch.cuda.amp.autocast(enabled=apex):
            if criterion is None:
                try:
                    preds = model(inputs, labels=labels)
                except ValueError:
                    preds = model(**inputs, labels=labels)
            else:
                try:
                    preds = model(inputs=inputs)
                except ValueError:
                    preds = model(**inputs)
            if isinstance(preds, dict) and "loss" in preds:
                loss = preds["loss"]
            else:
                if isinstance(inputs, dict):
                    if isinstance(preds, dict):
                        logits = preds['logits']
                    elif isinstance(preds, (list, tuple)):
                        logits = preds[0]
                    else:
                        logits = preds

                    if 'weights' in inputs:
                        loss = criterion(logits, labels.float(), inputs['weights'])
                    else:
                        loss = criterion(logits, labels.float())
                else:
                    loss = criterion(preds[0], preds[1])

        if teacher:
            with torch.no_grad():
                teacher_output = teacher(inputs)
                loss_distill = teacher_loss(preds, teacher_output)
                loss = loss_distill

        if gradient_accumulation_steps > 1:
            loss = loss / gradient_accumulation_steps

        losses.update(loss.item(), batch_size)
        scaler.scale(loss).backward()

        # if isinstance(preds, dict):
        #     acc = (preds["feature"].softmax(dim=1) * (labels > 0)).sum(axis=1).mean()
        # else:
        #     acc = (preds.softmax(dim=1) * (labels > 0)).sum(axis=1).mean()
        # acc = acc.detach().cpu().item()
        # acc_list.append(acc)

        # ---------------------fgm-------------
        if fgm:
            # attach the embedding
            fgm.attack(epsilon=1.0)
            with torch.cuda.amp.autocast(enabled=apex):
                preds = model(inputs)
                if isinstance(preds, dict) and "loss" in preds:
                    loss_avg = preds["loss"]
                else:
                    loss_avg = criterion(preds, labels)
            if gradient_accumulation_steps > 1:
                loss_avg = loss_avg / gradient_accumulation_steps
            losses.update(loss_avg.item(), batch_size)
            scaler.scale(loss_avg).backward()
            # restore the embedding
            fgm.restore()

        grad_norm = torch.nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)

        if (step + 1) % gradient_accumulation_steps == 0:
            scaler.step(optimizer)
            scaler.update()

            if ema_inst:
                ema_inst.update()

            optimizer.zero_grad()
            global_step += 1
            if batch_scheduler:
                batch_scheduler.step()

        if step % print_freq == 0 or (step + 1) == len(train_loader):
            print(
                "Epoch: [{0}][{1}/{2}] "
                "Elapsed {remain:s} "
                "Loss: {loss.val:.4f}({loss.avg:.4f}) "
                "Grad: {grad_norm:.4f}  "
                "LR: {lr:.8f}  "
                # "ACC: {acc: .4f}"
                .format(
                    epoch + 1,
                    step,
                    len(train_loader) - 1,
                    remain=timeSince(start, float(step + 1) / len(train_loader)),
                    loss=losses,
                    grad_norm=grad_norm,
                    lr=batch_scheduler.get_lr()[0] if batch_scheduler else optimizer.param_groups[0]["lr"],
                    # acc=np.mean(acc_list),
                )
            )

        if wandb and step % print_freq == 0:
            print({"[loss": losses.val, "[lr": optimizer.param_groups[0]["lr"]})

        if (step + 1) % save_step == 0 and epoch > -1:
            if ema_inst:
                ema_inst.apply_shadow()

            if ema_inst:
                ema_inst.restore()

    return losses.avg


@torch.no_grad()
def valid_fn(
    epoch,
    valid_loader,
    model,
    criterion,
    eval_metrics=None,
    apex: bool = False,
    device="cuda",
    data_collator=None,
    print_freq: int = 100,
):
    start = time.time()
    model.eval()
    losses = AverageMeter()
    pred_array = torch.tensor([], device=device)
    label_array = np.array([])

    for step, inputs in enumerate(valid_loader):
        if isinstance(inputs, (list, tuple)) and len(inputs) == 2:
            inputs, labels = inputs
        elif isinstance(inputs, dict):
            labels = inputs['labels']
            inputs.pop('labels', None)

        if data_collator:
            inputs = data_collator(inputs)

        for k, v in inputs.items():
            inputs[k] = v.to(device)
        labels = labels.to(device)

        batch_size = labels.size(0)
        with torch.cuda.amp.autocast(enabled=apex):
            try:
                preds = model(inputs, labels=labels)
            except ValueError:
                preds = model(inputs)
            else:
                logger.warning('inputs of model should be with or without labels')

            if isinstance(preds, dict) and "loss" in preds:
                loss = preds["loss"]
                pred = preds["output"]
            else:
                loss = criterion(preds, labels)
                pred = preds

        losses.update(loss.item(), batch_size)
        pred_array = torch.cat((pred_array, pred), dim=0)
        label_array = np.append(label_array, np.array(labels))

        if step % print_freq == 0 or step == (len(valid_loader) - 1):
            print(
                "Epoch: [{0}][{1}/{2}] "
                "Elapsed {remain:s} "
                "Loss: {loss.val:.4f}({loss.avg:.4f}) "
                # "ACC: {acc: .4f}"
                .format(
                    epoch + 1,
                    step,
                    len(valid_loader),
                    remain=timeSince(start, float(step + 1) / len(valid_loader)),
                    loss=losses,
                    # acc=np.mean(acc_list),
                )
            )

    if eval_metrics is not None:
        score = eval_metrics(label_array, pred_array)
        print(f"Epoch {epoch}, Score: {score}")
        return losses.avg, score

    return losses.avg


def inference_fn(test_loader, model, device):
    preds: List[np.ndarray] = []
    model.eval()
    model.to(device)
    tk0 = tqdm(test_loader, total=len(test_loader))
    for inputs in tk0:
        for k, v in inputs.items():
            inputs[k] = v.to(device)
        with torch.no_grad():
            y_preds = model(inputs)
        preds.append(y_preds.sigmoid().squeeze().to("cpu").numpy().reshape(-1))
    predictions = np.concatenate(preds)
    return predictions


class CustomTrainer(object):
    def __init__(self, model: Union[nn.Module], device: Optional[str] = None, apex: bool = False, teacher=None):
        if not device:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = device
        self.model = model
        self.apex = apex
        self.train_fn = train_fn
        self.valid_fn = valid_fn
        if teacher:
            self.teacher = teacher.eval()
            self.teacher_loss = torch.nn.MSELoss()

    def train(
        self,
        train_loader,
        optimizer,
        epochs=10,
        criterion=None,
        scheduler=None,
        valid_loader=None,
        eval_metrics=None,
        data_collator=None,
        max_grad_norm=10,
        use_fgm: bool = False,
        use_awp: bool = False,
        ema_decay: float = 0,
        dynamic_margin_fn: Optional[Callable] = None,
        gradient_checkpointing: bool = False,
        resume_from_checkpoint: Optional[Union[str, bool]] = None,
        **kwargs,
    ):
        logger.info('-------START TO TRAIN-------')
        if resume_from_checkpoint is not None:
            checkpoint = torch.load(resume_from_checkpoint)
            self.model.load_state_dict(checkpoint)

        if use_fgm:
            logger.info('[FGM] Use FGM adversarial')
            fgm = FGM(self.model.to(self.device))
        elif use_awp:
            logger.info('[AWP] Use AWP adversarial')
            awp = AWP(self.model.to(self.device), optimizer)

        if ema_decay > 0:
            logger.info('[EMA] Use EMA while training')
            ema_inst = EMA(self.model.to(self.device), ema_decay)
            ema_inst.register()

        if gradient_checkpointing:
            self.model.gradient_checkpointing_enable()

        for epoch in range(epochs):
            if dynamic_margin_fn:
                margin = dynamic_margin_fn(epoch)
                logger.info(f"[Dynamic margin] Epoch: {epoch}, Margin: {margin}")
                if criterion:
                    criterion.set_margin(margin)
                elif self.model.loss_fn is not None:
                    self.model.loss_fn.set_margin(margin)
                else:
                    raise ValueError(
                        'dynamic margin is True, but criterion of trainer and loss_fn of model are both none'
                    )

            # lr = scheduler(optimizer, epoch)

            # 放在这里, 是因为前面的set_margin也会产生新的参数位于cpu
            # 如果把参数直接放在cuda, 则导致parameter里没有出现该参数，无法分别设置学习率
            self.model = self.model.to(self.device)
            optimizer.zero_grad()
            self.train_fn(
                epoch=epoch,
                train_loader=train_loader,
                model=self.model,
                criterion=criterion,
                optimizer=optimizer,
                batch_scheduler=scheduler,
                fgm=fgm if use_fgm else None,
                awp=awp if use_awp else None,
                ema_inst=ema_inst if ema_decay > 0 else None,
                apex=self.apex,
                gradient_accumulation_steps=1,
                max_grad_norm=max_grad_norm,
                valid_loader=valid_loader,
                data_collator=data_collator,
                device=self.device,
            )

            if valid_loader is not None:
                self.valid_fn(
                    epoch=epoch,
                    valid_loader=valid_loader,
                    model=self.model,
                    criterion=criterion,
                    eval_metrics=eval_metrics,
                    apex=self.apex,
                    device=self.device,
                    data_collator=None,
                )

        torch.cuda.empty_cache()
        gc.collect()

    def predict(self, test_loader):
        self.model.eval()
        tbar = tqdm(test_loader)

        vectors = []
        with torch.no_grad():
            for idx, (inputs, labels) in enumerate(tbar):
                for k, v in inputs.items():
                    inputs[k] = v.to(self.device)

                vec = self.model(inputs)
                if isinstance(vec, dict):
                    vec = vec["embeddings"]
                vectors.append(vec.detach().cpu().numpy())

        V = np.concatenate(vectors)
        return V

    def save_state(self, output_dir: Optional[str] = None):
        # TODO: remove the loss_fn while save weights
        if not output_dir:
            output_dir = "./model.pth"
        torch.save(self.model.state_dict(), output_dir)

    def save_model(self, output_dir: Optional[str] = None, state_dict=None):
        pass


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def asMinutes(s):
    m = math.floor(s / 60)
    s = s - m * 60
    return "%dm %ds" % (m, s)


def timeSince(since, percent):
    now = time.time()
    s = now - since
    es = s / (percent)
    rs = es - s
    return "%s (remain %s)" % (asMinutes(s), asMinutes(rs))
